# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Unit tests for parallel task tree algorithm."""

from __future__ import (
    annotations,
)
import logging
import os
import random
import string
import threading
import time
import unittest
from concurrent.futures import (
    Future,
    ProcessPoolExecutor,
)
from functools import (
    partial,
)
from typing import (
    Any,
)
from unittest.mock import (
    MagicMock,
)

from bzfs_main.parallel_tasktree import (
    BARRIER_CHAR,
    CompletionCallback,
    CompletionCallbackResult,
    _build_dataset_tree,
    _complete_job_with_barriers,
    _make_tree_node,
    _Tree,
    _TreeNode,
    process_datasets_in_parallel,
)
from bzfs_main.utils import (
    SortedInterner,
    SynchronousExecutor,
)


#############################################################################
def suite() -> unittest.TestSuite:
    test_cases = [
        TestBuildTree,
        TestProcessDatasetsInParallel,
        TestCustomPriorityOrder,
        TestBarriersCleared,
        TestProcessPoolExecutor,
        TestParallelTaskTreeBenchmark,
    ]
    return unittest.TestSuite(unittest.TestLoader().loadTestsFromTestCase(test_case) for test_case in test_cases)


#############################################################################
class TestBuildTree(unittest.TestCase):

    def assert_keys_sorted(self, tree: dict[str, Any]) -> None:
        keys = list(tree.keys())
        self.assertEqual(sorted(keys), keys, f"Keys are not sorted: {keys}")
        for value in tree.values():
            if isinstance(value, dict):
                self.assert_keys_sorted(value)

    def test_basic_tree(self) -> None:
        datasets = ["pool", "pool/dataset", "pool/dataset/sub", "pool/other", "pool/other/sub/child"]
        expected_tree = {"pool": {"dataset": {"sub": {}}, "other": {"sub": {"child": {}}}}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_empty_input(self) -> None:
        datasets: list[str] = []
        expected_tree: _Tree = {}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)

    def test_single_root(self) -> None:
        datasets: list[str] = ["pool"]
        expected_tree: _Tree = {"pool": {}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_single_branch(self) -> None:
        datasets: list[str] = ["pool/dataset/sub/child"]
        expected_tree: _Tree = {"pool": {"dataset": {"sub": {"child": {}}}}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_multiple_roots(self) -> None:
        datasets: list[str] = ["pool", "otherpool", "anotherpool"]
        expected_tree: _Tree = {"anotherpool": {}, "otherpool": {}, "pool": {}}
        tree = _build_dataset_tree(sorted(datasets))
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_large_dataset(self) -> None:
        datasets: list[str] = [f"pool/dataset{i}" for i in range(100)]
        tree = _build_dataset_tree(sorted(datasets))
        self.assertEqual(100, len(tree["pool"]))
        self.assert_keys_sorted(tree)

    def test_nested_structure(self) -> None:
        datasets: list[str] = [
            "pool/parent",
            "pool/parent/child1",
            "pool/parent/child2",
            "pool/parent/child2/grandchild",
            "pool/parent/child3",
        ]
        expected_tree: _Tree = {"pool": {"parent": {"child1": {}, "child2": {"grandchild": {}}, "child3": {}}}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_no_children(self) -> None:
        datasets: list[str] = ["pool", "otherpool"]
        expected_tree: _Tree = {"otherpool": {}, "pool": {}}
        tree = _build_dataset_tree(sorted(datasets))
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_single_level(self) -> None:
        datasets: list[str] = ["pool", "pool1", "pool2", "pool3"]
        expected_tree: _Tree = {"pool": {}, "pool1": {}, "pool2": {}, "pool3": {}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_multiple_roots_with_hierarchy(self) -> None:
        datasets: list[str] = ["pool", "pool1", "pool1/dataset1", "pool2", "pool2/dataset2", "pool2/dataset2/sub", "pool3"]
        expected_tree: _Tree = {"pool": {}, "pool1": {"dataset1": {}}, "pool2": {"dataset2": {"sub": {}}}, "pool3": {}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_multiple_roots_flat(self) -> None:
        datasets: list[str] = ["root1", "root2", "root3", "root4"]
        expected_tree: _Tree = {"root1": {}, "root2": {}, "root3": {}, "root4": {}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_multiple_roots_mixed_depth(self) -> None:
        datasets: list[str] = ["a", "a/b", "a/b/c", "x", "x/y", "z", "z/1", "z/2", "z/2/3"]
        expected_tree: _Tree = {"a": {"b": {"c": {}}}, "x": {"y": {}}, "z": {"1": {}, "2": {"3": {}}}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_tree_with_missing_intermediate_nodes(self) -> None:
        datasets: list[str] = ["a", "a/b/c", "z/2/3"]
        expected_tree: _Tree = {"a": {"b": {"c": {}}}, "z": {"2": {"3": {}}}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)

    def test_tree_with_barriers(self) -> None:
        br = BARRIER_CHAR
        datasets: list[str] = [
            "a/b/c",
            "a/b/c/0d",
            "a/b/c/1d",
            f"a/b/c/{br}/prune",
            f"a/b/c/{br}/prune/monitor",
            f"a/b/c/{br}/{br}/done",
        ]
        expected_tree: _Tree = {"a": {"b": {"c": {"0d": {}, "1d": {}, br: {"prune": {"monitor": {}}, br: {"done": {}}}}}}}
        tree = _build_dataset_tree(datasets)
        self.assertEqual(expected_tree, tree)
        self.assert_keys_sorted(tree)


#############################################################################
class TestProcessDatasetsInParallel(unittest.TestCase):

    def test_str_treenode(self) -> None:
        self.assertTrue(bool(str(_make_tree_node("foo", "foo", {}))))

    def test_termination_event_pre_set_stops_before_submission(self) -> None:
        """If termination_event is set before scheduling, no tasks are submitted and the run fails."""
        log = MagicMock(logging.Logger)
        datasets = ["a", "b", "c"]
        calls: list[str] = []

        def process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
            calls.append(dataset)

            def _completion_callback(todo_futures: set[Future[CompletionCallback]]) -> CompletionCallbackResult:
                return CompletionCallbackResult(no_skip=True, fail=False)

            return _completion_callback

        termination_event = threading.Event()
        termination_event.set()

        failed = process_datasets_in_parallel(
            log=log,
            datasets=datasets,
            process_dataset=process_dataset,
            max_workers=2,
            termination_event=termination_event,
            enable_barriers=False,
            is_test_mode=True,
        )

        self.assertTrue(failed, "Termination should mark the run as failed")
        self.assertEqual(0, len(calls), "No dataset should be submitted when already terminated")

    def test_termination_event_set_during_sleep_stops_new_submissions(self) -> None:
        """When termination_event is set while the coordinator sleeps, it should wake early, stop submitting, and fail."""

        log = MagicMock(logging.Logger)
        datasets = ["a", "b", "c"]
        calls: list[str] = []
        termination_event = threading.Event()

        # Ensure the scheduler sleeps between submissions to hit the termination_event.wait() path
        def interval_nanos(last_update_nanos: int, dataset: str, submitted_count: int) -> int:
            # Large enough to allow the background thread to set the event and wake the wait early
            return 500_000_000  # 0.5s

        def process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
            calls.append(dataset)

            def _completion_callback(todo_futures: set[Future[CompletionCallback]]) -> CompletionCallbackResult:
                return CompletionCallbackResult(no_skip=True, fail=False)

            return _completion_callback

        # Background thread that sets termination once the first task has been submitted
        def trigger_termination() -> None:
            # Busy-wait is fine here because it runs for a very short time and simplifies determinism
            while len(calls) < 1:
                time.sleep(0.005)
            # Give the coordinator a moment to enter wait(); then signal termination to wake it
            time.sleep(0.01)
            termination_event.set()

        t = threading.Thread(target=trigger_termination)
        t.start()

        failed = process_datasets_in_parallel(
            log=log,
            datasets=datasets,
            process_dataset=process_dataset,
            max_workers=2,
            interval_nanos=interval_nanos,
            termination_event=termination_event,
            enable_barriers=False,
            is_test_mode=True,
        )

        t.join(timeout=2)

        self.assertTrue(failed, "Termination should mark the run as failed")
        # Exactly one submission is expected: first submitted before sleep, then termination prevents further submissions
        self.assertEqual(1, len(calls), f"Expected 1 submitted dataset, got {len(calls)}: {calls}")

    def test_explicit_sync_executor_runs_inline_on_main_thread(self) -> None:
        """With an explicit SynchronousExecutor, tasks execute inline on the main thread.

        Uses a dataset list with siblings to prove that even with ``max_workers>1`` and parallelizable structure, an explicit
        synchronous executor forces in-thread execution and preserves deterministic order.
        """

        log = MagicMock(logging.Logger)
        datasets = ["a", "a/b", "a/c"]  # siblings under 'a'

        lock: threading.Lock = threading.Lock()
        main_ident: int = threading.get_ident()
        calls: list[tuple[str, int]] = []

        def process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
            with lock:
                calls.append((dataset, threading.get_ident()))

            def _completion_callback(todo_futures: set[Future[CompletionCallback]]) -> CompletionCallbackResult:
                return CompletionCallbackResult(no_skip=True, fail=False)

            return _completion_callback

        failed = process_datasets_in_parallel(
            log=log,
            datasets=datasets,
            process_dataset=process_dataset,
            max_workers=4,
            executor=SynchronousExecutor(),
            enable_barriers=False,
            is_test_mode=True,
        )

        self.assertFalse(failed)
        # Expected order: root 'a' first, then its children in lexicographic order
        self.assertEqual([d for d, _ in calls], ["a", "a/b", "a/c"])
        self.assertTrue(all(tid == main_ident for _, tid in calls), msg=f"Calls not on main thread: {calls}")


#############################################################################
class TestCustomPriorityOrder(unittest.TestCase):

    def test_custom_priority_orders_available_datasets(self) -> None:
        """Custom priority uses integer "size" (cost) per dataset to decide order among available datasets.

        Datasets with the largest size must be processed first, while always respecting the dependency rule that a parent
        must complete before any of its children can be processed. Two roots ("r" and "s") with children are used. With
        max_workers=1 the order is deterministic.
        """
        log = MagicMock(logging.Logger)

        # Sorted input list with two roots and children; no barriers involved
        datasets = ["r", "r/a", "r/b", "r/c", "s", "s/x", "s/y"]

        # Simulated dataset sizes (aka cost). Largest size should be picked first among available datasets.
        sizes: dict[str, int] = {
            "r": 10,
            "r/a": 30,
            "r/b": 20,
            "r/c": 40,
            "s": 50,
            "s/x": 5,
            "s/y": 60,
        }

        # Priority: smaller compares first; use negative size for max-heap behavior, add dataset for stable tie-break
        def priority(ds: str) -> tuple[int, str]:
            return (-sizes[ds], ds)

        calls: list[str] = []

        def process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
            calls.append(dataset)

            def _completion_callback(todo_futures: set[Future[CompletionCallback]]) -> CompletionCallbackResult:
                return CompletionCallbackResult(no_skip=True, fail=False)

            return _completion_callback

        failed = process_datasets_in_parallel(
            log=log,
            datasets=datasets,
            process_dataset=process_dataset,
            priority=priority,
            max_workers=1,
            enable_barriers=False,
            is_test_mode=True,
        )

        self.assertFalse(failed)
        # Expected order determined by sizes while respecting dependencies:
        # Roots available initially: pick 's'(50) over 'r'(10); after 's', its children become available along with 'r'.
        # Pick 's/y'(60) over 'r'(10) and 's/x'(5); then 'r'(10) before the remaining 's/x'(5). After 'r', pick 'r/c'(40),
        # 'r/a'(30), 'r/b'(20), then last remaining 's/x'(5).
        expected = ["s", "s/y", "r", "r/c", "r/a", "r/b", "s/x"]
        self.assertEqual(expected, calls)


def make_tree_node(dataset: str, children: _Tree, parent: _TreeNode | None = None) -> _TreeNode:
    return _make_tree_node(priority=dataset, dataset=dataset, children=children, parent=parent)


#############################################################################
class TestBarriersCleared(unittest.TestCase):

    def test_failure_clears_ancestor_barriers_and_sets_empty_barrier(self) -> None:
        """Verifies that a first failure clears barriers for the node and all its ancestors and sets empty_barrier.

        This exercises the ancestor-walking guard by confirming barriers_cleared flags are set along the chain, and that
        barriers are set to the empty_barrier on each visited node when handling a failure without further propagation
        (pending > 0 suppresses the subsequent while-loop).
        """

        # Build a -> b -> c chain
        a = make_tree_node("a", {})
        b = make_tree_node("a/b", {}, parent=a)
        c = make_tree_node("a/b/c", {}, parent=b)

        # Prevent the completion-propagation while-loop from running to keep the test focused on the failure path only
        c.mut.pending = 1
        a.mut.pending = 0
        b.mut.pending = 0

        priority_queue: list = []
        datasets_set: SortedInterner[str] = SortedInterner([])
        empty_barrier = make_tree_node("empty_barrier", {})

        # First failure at deepest node
        _complete_job_with_barriers(
            c,
            no_skip=False,
            priority_queue=priority_queue,
            priority=lambda dataset: dataset,
            datasets_set=datasets_set,
            empty_barrier=empty_barrier,
        )

        # Check that the node and its ancestors have barriers cleared and point to the empty_barrier
        self.assertTrue(c.mut.barriers_cleared)
        self.assertTrue(b.mut.barriers_cleared)
        self.assertTrue(a.mut.barriers_cleared)
        self.assertIs(c.mut.barrier, empty_barrier)
        self.assertIs(b.mut.barrier, empty_barrier)
        self.assertIs(a.mut.barrier, empty_barrier)

    def test_ancestor_walking_stops_at_cleared_ancestor(self) -> None:
        """Second failure in the same subtree should stop clearing at the first ancestor with barriers_cleared set.

        We confirm by setting a custom barrier object on an ancestor and ensuring it remains unchanged after the second
        failure. We again keep pending > 0 to avoid the subsequent completion-propagation loop from running.
        """

        # Build a -> b -> c
        a = make_tree_node("a", {})
        b = make_tree_node("a/b", {}, parent=a)
        c = make_tree_node("a/b/c", {}, parent=b)

        # First failure clears barriers up to root
        c.mut.pending = 1  # suppress while-loop
        priority_queue: list = []
        datasets_set: SortedInterner[str] = SortedInterner([])
        empty_barrier = make_tree_node("empty_barrier", {})
        _complete_job_with_barriers(
            c,
            no_skip=False,
            priority_queue=priority_queue,
            priority=lambda dataset: dataset,
            datasets_set=datasets_set,
            empty_barrier=empty_barrier,
        )

        # Verify barriers cleared
        self.assertTrue(a.mut.barriers_cleared)
        self.assertTrue(b.mut.barriers_cleared)
        self.assertTrue(c.mut.barriers_cleared)

        # Place a custom marker on ancestor 'a' to detect unwanted overwrites; ancestor walking must stop at 'b'.
        marker = make_tree_node("custom_marker", {})
        a.mut.barrier = marker

        # Now fail deeper sibling 'd' under 'b' and ensure 'a' stays untouched by the barrier-clearing loop
        d = make_tree_node("a/b/d", {}, parent=b)
        d.mut.pending = 1  # suppress while-loop
        _complete_job_with_barriers(
            d,
            no_skip=False,
            priority_queue=priority_queue,
            priority=lambda dataset: dataset,
            datasets_set=datasets_set,
            empty_barrier=empty_barrier,
        )

        # 'd' gets barriers cleared; 'b' and 'a' remain with barriers cleared but 'a' barrier should still be the custom marker
        self.assertTrue(d.mut.barriers_cleared)
        self.assertTrue(b.mut.barriers_cleared)
        self.assertTrue(a.mut.barriers_cleared)
        self.assertIs(
            a.mut.barrier, marker, "Ancestor walking should stop at first cleared ancestor and not touch higher ancestors"
        )

    def test_early_break_can_open_ancestor_barrier_after_failure(self) -> None:
        """Demonstrate that breaking when encountering an empty_barrier on an intermediate ancestor would allow a higher
        ancestor barrier to open after a failure deeper in the tree; This exposes that an early-break optimization would be
        unsafe unless all higher ancestors are already set to the empty_barrier."""

        log = MagicMock(logging.Logger)
        br = BARRIER_CHAR

        # Tree under 'x':
        # - x/node/child -> success, which opens x/node barrier and enqueues x/node/~/bar/fail
        # - x/node/~/bar/fail -> failure (no_skip=False)
        # - x/other -> delayed completion to keep x.pending > 0 at the time of failure
        # - x/~/after -> barrier job at ancestor 'x' that must NOT start if any descendant fails
        datasets = [
            "x",
            "x/node",
            "x/node/child",
            f"x/node/{br}/bar/fail",
            "x/other",
            f"x/{br}/after",
        ]
        self.assertEqual(sorted(datasets), datasets)

        calls: list[str] = []
        lock = threading.Lock()
        failure_done = threading.Event()

        def record(dataset: str) -> None:
            with lock:
                calls.append(dataset)

        def process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
            # Simulate long-running sibling to keep ancestor 'x' pending > 0 during failure handling
            if dataset == "x/other":
                time.sleep(0.05)  # give others a head start
                failure_done.wait(timeout=2)  # finish only after failure has been handled

            record(dataset)

            def _completion_callback(todo_futures: set[Future[CompletionCallback]]) -> CompletionCallbackResult:
                if dataset == f"x/node/{br}/bar/fail":
                    # Signal that failure handling has run; return no_skip=False (skip subtree) but not fail the run
                    failure_done.set()
                    return CompletionCallbackResult(no_skip=False, fail=False)
                return CompletionCallbackResult(no_skip=True, fail=False)

            return _completion_callback

        failed = process_datasets_in_parallel(
            log=log,
            datasets=datasets,
            process_dataset=process_dataset,
            max_workers=2,
            enable_barriers=True,
            is_test_mode=True,
        )

        # The run should not be marked failed since fail=False, but the ancestor barrier 'x/~/after' must NOT start.
        # If it did start, the early-break allowed opening of ancestor barrier after a failure.
        self.assertFalse(failed)
        self.assertNotIn(f"x/{br}/after", calls, msg=f"Ancestor barrier job should not start after failure, calls={calls}")


#############################################################################
# Top-level helpers used by TestProcessPoolExecutor pickling tests.
#
# Process pools require pickleable, importable callables. We return a functools.partial of a top-level function to carry
# simple picklable state.
PP_CALLS: list[tuple[str, int, int]] = []  # (dataset, worker_pid, main_pid_at_callback)


def pp_completion_callback(
    dataset: str, worker_pid: int, todo_futures: set[Future[CompletionCallback]]
) -> CompletionCallbackResult:
    """Top-level callback for ProcessPoolExecutor tests; appends a record and continues.

    Must remain at module scope to be pickleable for functools.partial().
    """
    PP_CALLS.append((dataset, worker_pid, os.getpid()))
    return CompletionCallbackResult(no_skip=True, fail=False)


def pp_process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
    """Top-level worker used by process pool tests; returns a pickleable callback.

    Binds only simple values (dataset, worker_pid) into the partial so unpickling is reliable across processes.
    """
    worker_pid: int = os.getpid()
    return partial(pp_completion_callback, dataset, worker_pid)


#############################################################################
class TestProcessPoolExecutor(unittest.TestCase):

    def test_process_pool_executor_with_picklable_callback_via_partial(self) -> None:
        """ProcessPoolExecutor works when process_dataset returns a pickleable callback (via partial).

        - Ensures the returned CompletionCallback is reconstructed in the coordinator process and can record results.
        - Confirms that workers ran in separate processes by comparing PIDs.
        - Validates dependency ordering: the root runs before its children; sibling completion order is not asserted.
        """

        # Reset global call log
        PP_CALLS.clear()

        log = MagicMock(logging.Logger)
        datasets = ["a", "a/b", "a/c"]  # siblings under 'a'
        main_pid = os.getpid()

        failed = process_datasets_in_parallel(
            log=log,
            datasets=datasets,
            process_dataset=pp_process_dataset,
            max_workers=2,
            executor=ProcessPoolExecutor(max_workers=2),
            enable_barriers=False,
            is_test_mode=True,
        )

        self.assertFalse(failed)
        self.assertEqual(3, len(PP_CALLS), PP_CALLS)
        # Root must complete before children are scheduled
        self.assertEqual("a", PP_CALLS[0][0], PP_CALLS)
        # Sibling order may vary; assert set equality for the remaining two
        remaining = [rec[0] for rec in PP_CALLS[1:]]
        self.assertEqual({"a/b", "a/c"}, set(remaining), PP_CALLS)
        # At least one worker must be a separate process
        self.assertTrue(any(worker_pid != main_pid for _, worker_pid, _ in PP_CALLS), PP_CALLS)
        # Callback runs in the main process
        self.assertTrue(all(main_pid == cb_main_pid for _, _, cb_main_pid in PP_CALLS), PP_CALLS)


#############################################################################
class TestParallelTaskTreeBenchmark(unittest.TestCase):

    @staticmethod
    def generate_unique_datasets(num_datasets: int, length: int) -> list[str]:
        # Create a realistic hierarchy to test tree-building performance
        # Example: tank/group_xxxx/host_yyyy/data_zzzzzzzzzzzz...
        datasets: set[str] = set()

        # Characters for random components
        chars = string.ascii_lowercase + string.digits

        while len(datasets) < num_datasets:
            # Generate components to ensure some path sharing
            l1 = "tank"
            l2 = f"group_{''.join(random.choices(chars, k=4))}"
            l3 = f"host_{''.join(random.choices(chars, k=4))}"

            # Ensure the final component is long enough to meet the total length
            prefix = f"{l1}/{l2}/{l3}/"
            remaining_len = length - len(prefix)
            if remaining_len <= 0:
                raise ValueError("Target length is too short for the fixed prefix.")

            leaf = "".join(random.choices(chars, k=remaining_len))
            dataset = prefix + leaf
            datasets.add(dataset)

        return sorted(datasets)

    def _run_benchmark(self, num_datasets: int, enable_barriers: bool, max_workers: int = 2 * (os.cpu_count() or 1)) -> None:

        def dummy_process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
            """A dummy function that does nothing, to benchmark the framework overhead."""

            def _completion_callback(todo_futures: set[Future[CompletionCallback]]) -> CompletionCallbackResult:
                """A dummy function that never skips."""
                return CompletionCallbackResult(no_skip=True, fail=False)

            return _completion_callback

        log = logging.getLogger("TestParallelTaskTreeBenchmark")
        log.setLevel(logging.INFO)
        if not log.handlers:
            log.addHandler(logging.StreamHandler())

        datasets = self.generate_unique_datasets(num_datasets=num_datasets, length=100)

        import gc

        gc.collect()
        start_time = time.monotonic()

        failed = process_datasets_in_parallel(
            log=log,
            datasets=datasets,
            process_dataset=dummy_process_dataset,
            max_workers=max_workers,
            enable_barriers=enable_barriers,
            is_test_mode=False,
        )

        end_time = time.monotonic()
        elapsed_time = end_time - start_time
        throughput = round(num_datasets / elapsed_time)
        log.info("=================================================")
        log.info(f"Results for datasets={num_datasets}, enable_barriers={enable_barriers}, max_workers={max_workers} ...")
        log.info(f"Total elapsed time: {elapsed_time:.2f} seconds")
        log.info(f"Throughput: {throughput} datasets/second")
        self.assertFalse(failed, "The process should not report failure.")

    def test_benchmark_10k_datasets(self) -> None:
        for enable_barriers in [False, True]:
            self._run_benchmark(num_datasets=10_000, enable_barriers=enable_barriers)

    @unittest.skip("benchmark; enable for performance comparison")
    def test_benchmark_100k_datasets(self) -> None:
        for enable_barriers in [False, True]:
            self._run_benchmark(num_datasets=100_000, enable_barriers=enable_barriers)

    @unittest.skip("benchmark; enable for performance comparison")
    def test_benchmark_1m_datasets(self) -> None:
        for enable_barriers in [False, True]:
            self._run_benchmark(num_datasets=1_000_000, enable_barriers=enable_barriers)



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Policy layer for the generic parallel task tree scheduling algorithm.

Purpose: Provide bzfs-specific behavior on top of the policy-free generic ``parallel_tasktree`` scheduling algorithm:
retries, skip-on-error modes (fail/dataset/tree), and logging.

Assumptions: Callers provide a thread-safe ``process_dataset(dataset, tid, Retry) -> bool``. Dataset list is sorted and
unique (enforced by tests).

Design rationale: Keep scheduling generic and reusable while concentrating error handling and side-effects here. This module
exposes a stable API for callers like ``bzfs`` and ``bzfs_jobrunner``.
"""

from __future__ import (
    annotations,
)
import logging
import os
import subprocess
import threading
import time
from concurrent.futures import (
    Future,
)
from logging import (
    Logger,
)
from typing import (
    Callable,
)

from bzfs_main.parallel_tasktree import (
    CompletionCallback,
    CompletionCallbackResult,
    process_datasets_in_parallel,
)
from bzfs_main.retry import (
    Retry,
    RetryPolicy,
    run_with_retries,
)
from bzfs_main.utils import (
    dry,
    human_readable_duration,
)


def process_datasets_in_parallel_and_fault_tolerant(
    log: Logger,
    datasets: list[str],  # (sorted) list of datasets to process
    process_dataset: Callable[
        [str, str, Retry], bool  # lambda: dataset, tid, Retry; return False to skip subtree; must be thread-safe
    ],
    skip_tree_on_error: Callable[[str], bool],  # lambda: dataset # called on error; return True to skip subtree on error
    skip_on_error: str = "fail",
    max_workers: int = os.cpu_count() or 1,
    interval_nanos: Callable[
        [int, str, int], int
    ] = lambda last_update_nanos, dataset, submitted_count: 0,  # optionally spread tasks out over time; e.g. for jitter
    termination_event: threading.Event | None = None,  # optional event to request early async termination
    termination_handler: Callable[[], None] = lambda: None,
    task_name: str = "Task",
    enable_barriers: bool | None = None,  # for testing only; None means 'auto-detect'
    append_exception: Callable[[BaseException, str, str], None] = lambda ex, task, dataset: None,  # called on nonfatal error
    retry_policy: RetryPolicy | None = None,
    dry_run: bool = False,
    is_test_mode: bool = False,
) -> bool:  # returns True if any dataset processing failed, False if all succeeded
    """Runs datasets in parallel with retries and skip policy.

    Purpose: Adapt the generic engine to bzfs needs by wrapping the worker function with retries and determining skip/fail
    behavior on completion.

    Assumptions: ``skip_on_error`` is one of {"fail","dataset","tree"}. ``skip_tree_on_error(dataset)`` returns True if
    subtree should be skipped.

    Design rationale: The completion callback runs in the main thread, enabling safe cancellation of in-flight futures for
    fail-fast while keeping worker threads free of policy decisions.
    """
    assert callable(process_dataset)
    assert callable(skip_tree_on_error)
    termination_event = threading.Event() if termination_event is None else termination_event
    assert "%" not in task_name
    assert callable(append_exception)
    retry_policy = RetryPolicy.no_retries() if retry_policy is None else retry_policy
    len_datasets: int = len(datasets)
    is_debug: bool = log.isEnabledFor(logging.DEBUG)

    def _process_dataset(dataset: str, submitted_count: int) -> CompletionCallback:
        """Wrapper around process_dataset(); adds retries plus a callback determining if to fail or skip subtree on error."""
        tid: str = f"{submitted_count}/{len_datasets}"
        start_time_nanos: int = time.monotonic_ns()
        exception = None
        no_skip: bool = False
        try:
            no_skip = run_with_retries(log, retry_policy, termination_event, process_dataset, dataset, tid)
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired, SystemExit, UnicodeDecodeError) as e:
            exception = e  # may be reraised later
        finally:
            if is_debug:
                elapsed_duration: str = human_readable_duration(time.monotonic_ns() - start_time_nanos)
                log.debug(dry(f"{tid} {task_name} done: %s took %s", dry_run), dataset, elapsed_duration)

        def _completion_callback(todo_futures: set[Future[CompletionCallback]]) -> CompletionCallbackResult:
            """CompletionCallback determining if to fail or skip subtree on error; Runs in the (single) main thread as part
            of the coordination loop."""
            nonlocal no_skip
            fail: bool = False
            if exception is not None:
                fail = True
                if skip_on_error == "fail" or termination_event.is_set():
                    for todo_future in todo_futures:
                        todo_future.cancel()
                    termination_handler()
                    raise exception
                no_skip = not (skip_on_error == "tree" or skip_tree_on_error(dataset))
                log.error("%s", exception)
                append_exception(exception, task_name, dataset)
            return CompletionCallbackResult(no_skip=no_skip, fail=fail)

        return _completion_callback

    return process_datasets_in_parallel(
        log=log,
        datasets=datasets,
        process_dataset=_process_dataset,
        max_workers=max_workers,
        interval_nanos=interval_nanos,
        termination_event=termination_event,
        enable_barriers=enable_barriers,
        is_test_mode=is_test_mode,
    )



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The core replication algorithm is in replicate_dataset(), which performs reliable full and/or incremental 'zfs send' and
'zfs receive' operations on snapshots, using resumable ZFS sends when possible.

For replication of multiple datasets, including recursive replication, see bzfs.py/replicate_datasets().
"""

from __future__ import (
    annotations,
)
import os
import re
import shlex
import subprocess
import sys
import threading
import time
from collections.abc import (
    Iterable,
    Iterator,
)
from concurrent.futures import (
    Executor,
    Future,
)
from subprocess import (
    DEVNULL,
    PIPE,
)
from typing import (
    TYPE_CHECKING,
    Final,
)

from bzfs_main.argparse_actions import (
    has_timerange_filter,
)
from bzfs_main.connection import (
    DEDICATED,
    SHARED,
    ConnectionPool,
    maybe_inject_error,
    refresh_ssh_connection_if_necessary,
    run_ssh_command,
    timeout,
    try_ssh_command,
)
from bzfs_main.detect import (
    ZFS_VERSION_IS_AT_LEAST_2_1_0,
    ZFS_VERSION_IS_AT_LEAST_2_2_0,
    are_bookmarks_enabled,
    is_zpool_feature_enabled_or_active,
)
from bzfs_main.filter import (
    filter_properties,
    filter_snapshots,
)
from bzfs_main.incremental_send_steps import (
    incremental_send_steps,
)
from bzfs_main.parallel_batch_cmd import (
    run_ssh_cmd_batched,
    run_ssh_cmd_parallel,
)
from bzfs_main.parallel_iterator import (
    parallel_iterator,
    run_in_parallel,
)
from bzfs_main.progress_reporter import (
    PV_FILE_THREAD_SEPARATOR,
)
from bzfs_main.retry import (
    Retry,
    RetryableError,
)
from bzfs_main.utils import (
    DONT_SKIP_DATASET,
    FILE_PERMISSIONS,
    LOG_DEBUG,
    LOG_TRACE,
    Subprocesses,
    append_if_absent,
    cut,
    die,
    getenv_bool,
    human_readable_bytes,
    is_descendant,
    list_formatter,
    open_nofollow,
    replace_prefix,
    stderr_to_str,
    xprint,
)

if TYPE_CHECKING:  # pragma: no cover - for type hints only
    from bzfs_main.bzfs import (
        Job,
    )
    from bzfs_main.configuration import (
        Params,
        Remote,
    )


# constants:
INJECT_DST_PIPE_FAIL_KBYTES: Final[int] = 400  # for testing only
RIGHT_JUST: Final[int] = 7


def replicate_dataset(job: Job, src_dataset: str, tid: str, retry: Retry) -> bool:
    """Replicates src_dataset to dst_dataset (thread-safe); For replication of multiple datasets, including recursive
    replication, see bzfs.py/replicate_datasets()."""
    p, log = job.params, job.params.log
    src, dst = p.src, p.dst
    retry_count: int = retry.count
    dst_dataset: str = replace_prefix(src_dataset, old_prefix=src.root_dataset, new_prefix=dst.root_dataset)
    log.debug(p.dry(f"{tid} Replicating: %s"), f"{src_dataset} --> {dst_dataset} ...")

    list_result: bool | tuple[list[str], list[str], list[str], set[str], str, str] = _list_and_filter_src_and_dst_snapshots(
        job, src_dataset, dst_dataset
    )
    if isinstance(list_result, bool):
        return list_result
    (
        basis_src_snapshots_with_guids,
        _src_snapshots_with_guids,
        dst_snapshots_with_guids,
        included_src_guids,
        latest_src_snapshot,
        oldest_src_snapshot,
    ) = list_result
    log.debug("latest_src_snapshot: %s", latest_src_snapshot)
    latest_dst_snapshot: str = ""
    latest_common_src_snapshot: str = ""
    done_checking: bool = False

    if job.dst_dataset_exists[dst_dataset]:
        rollback_result: bool | tuple[str, str, bool] = _rollback_dst_dataset_if_necessary(
            job,
            dst_dataset,
            latest_src_snapshot,
            basis_src_snapshots_with_guids,
            dst_snapshots_with_guids,
            done_checking,
            tid,
        )
        if isinstance(rollback_result, bool):
            return rollback_result
        latest_dst_snapshot, latest_common_src_snapshot, done_checking = rollback_result

    log.debug("latest_common_src_snapshot: %s", latest_common_src_snapshot)  # is a snapshot or bookmark
    log.log(LOG_TRACE, "latest_dst_snapshot: %s", latest_dst_snapshot)
    props_cache: dict[tuple[str, str, str], dict[str, str | None]] = {}  # fresh empty ZFS props cache for each dataset
    dry_run_no_send: bool = False
    if not latest_common_src_snapshot:
        # no common snapshot exists; delete all dst snapshots and perform a full send of the oldest selected src snapshot
        full_result: tuple[str, bool, bool, int] = _replicate_dataset_fully(
            job,
            src_dataset,
            dst_dataset,
            oldest_src_snapshot,
            latest_src_snapshot,
            latest_dst_snapshot,
            dst_snapshots_with_guids,
            props_cache,
            dry_run_no_send,
            done_checking,
            retry_count,
            tid,
        )
        # we have now created a common snapshot
        latest_common_src_snapshot, dry_run_no_send, done_checking, retry_count = full_result
    if latest_common_src_snapshot:
        # finally, incrementally replicate all selected snapshots from latest common snapshot until latest src snapshot
        _replicate_dataset_incrementally(
            job,
            src_dataset,
            dst_dataset,
            latest_common_src_snapshot,
            latest_src_snapshot,
            basis_src_snapshots_with_guids,
            included_src_guids,
            props_cache,
            dry_run_no_send,
            done_checking,
            retry_count,
            tid,
        )
    return True


def _list_and_filter_src_and_dst_snapshots(
    job: Job, src_dataset: str, dst_dataset: str
) -> bool | tuple[list[str], list[str], list[str], set[str], str, str]:
    """On replication, list and filter src and dst snapshots."""
    p, log = job.params, job.params.log
    src, dst = p.src, p.dst

    # list GUID and name for dst snapshots, sorted ascending by createtxg (more precise than creation time)
    dst_cmd: list[str] = p.split_args(f"{p.zfs_program} list -t snapshot -d 1 -s createtxg -Hp -o guid,name", dst_dataset)

    # list GUID and name for src snapshots + bookmarks, primarily sort ascending by transaction group (which is more
    # precise than creation time), secondarily sort such that snapshots appear after bookmarks for the same GUID.
    # Note: A snapshot and its ZFS bookmarks always have the same GUID, creation time and transaction group. A snapshot
    # changes its transaction group but retains its creation time and GUID on 'zfs receive' on another pool, i.e.
    # comparing createtxg is only meaningful within a single pool, not across pools from src to dst. Comparing creation
    # time remains meaningful across pools from src to dst. Creation time is a UTC Unix time in integer seconds.
    # Note that 'zfs create', 'zfs snapshot' and 'zfs bookmark' CLIs enforce that snapshot names must not contain a '#'
    # char, bookmark names must not contain a '@' char, and dataset names must not contain a '#' or '@' char.
    # GUID and creation time also do not contain a '#' or '@' char.
    filter_needs_creation_time: bool = has_timerange_filter(p.snapshot_filters)
    types: str = "snapshot,bookmark" if p.use_bookmark and are_bookmarks_enabled(p, src) else "snapshot"
    props: str = job.creation_prefix + "creation,guid,name" if filter_needs_creation_time else "guid,name"
    src_cmd = p.split_args(f"{p.zfs_program} list -t {types} -s createtxg -s type -d 1 -Hp -o {props}", src_dataset)
    job.maybe_inject_delete(src, dataset=src_dataset, delete_trigger="zfs_list_snapshot_src")
    src_snapshots_and_bookmarks, dst_snapshots_with_guids_str = run_in_parallel(  # list src+dst snapshots in parallel
        lambda: try_ssh_command(job, src, LOG_TRACE, cmd=src_cmd),
        lambda: try_ssh_command(job, dst, LOG_TRACE, cmd=dst_cmd, error_trigger="zfs_list_snapshot_dst"),
    )
    job.dst_dataset_exists[dst_dataset] = dst_snapshots_with_guids_str is not None
    dst_snapshots_with_guids: list[str] = (dst_snapshots_with_guids_str or "").splitlines()
    if src_snapshots_and_bookmarks is None:
        log.warning("Third party deleted source: %s", src_dataset)
        return False  # src dataset has been deleted by some third party while we're running - nothing to do anymore
    src_snapshots_with_guids: list[str] = src_snapshots_and_bookmarks.splitlines()
    src_snapshots_and_bookmarks = None
    if len(dst_snapshots_with_guids) == 0 and "bookmark" in types:
        # src bookmarks serve no purpose if the destination dataset has no snapshot; ignore them
        src_snapshots_with_guids = [snapshot for snapshot in src_snapshots_with_guids if "@" in snapshot]
    num_src_snapshots_found: int = sum(1 for snapshot in src_snapshots_with_guids if "@" in snapshot)
    with job.stats_lock:
        job.num_snapshots_found += num_src_snapshots_found
    # apply include/exclude regexes to ignore irrelevant src snapshots
    basis_src_snapshots_with_guids: list[str] = src_snapshots_with_guids
    src_snapshots_with_guids = filter_snapshots(job, src_snapshots_with_guids)
    if filter_needs_creation_time:
        src_snapshots_with_guids = cut(field=2, lines=src_snapshots_with_guids)
        basis_src_snapshots_with_guids = cut(field=2, lines=basis_src_snapshots_with_guids)

    # find oldest and latest "true" snapshot, as well as GUIDs of all snapshots and bookmarks.
    # a snapshot is "true" if it is not a bookmark.
    oldest_src_snapshot: str = ""
    latest_src_snapshot: str = ""
    included_src_guids: set[str] = set()
    for line in src_snapshots_with_guids:
        guid, snapshot = line.split("\t", 1)
        if "@" in snapshot:
            included_src_guids.add(guid)
            latest_src_snapshot = snapshot
            if not oldest_src_snapshot:
                oldest_src_snapshot = snapshot
    if len(src_snapshots_with_guids) == 0:
        if p.skip_missing_snapshots == "fail":
            die(f"Source dataset includes no snapshot: {src_dataset}. Consider using --skip-missing-snapshots=dataset")
        elif p.skip_missing_snapshots == "dataset":
            log.warning("Skipping source dataset because it includes no snapshot: %s", src_dataset)
            if p.recursive and not job.dst_dataset_exists[dst_dataset]:
                log.warning("Also skipping descendant datasets as dst dataset does not exist for %s", src_dataset)
            return job.dst_dataset_exists[dst_dataset]
    return (
        basis_src_snapshots_with_guids,
        src_snapshots_with_guids,
        dst_snapshots_with_guids,
        included_src_guids,
        latest_src_snapshot,
        oldest_src_snapshot,
    )


def _rollback_dst_dataset_if_necessary(
    job: Job,
    dst_dataset: str,
    latest_src_snapshot: str,
    src_snapshots_with_guids: list[str],
    dst_snapshots_with_guids: list[str],
    done_checking: bool,
    tid: str,
) -> bool | tuple[str, str, bool]:
    """On replication, rollback dst if necessary."""
    p, log = job.params, job.params.log
    dst = p.dst
    latest_dst_snapshot: str = ""
    latest_dst_guid: str = ""
    if len(dst_snapshots_with_guids) > 0:
        latest_dst_guid, latest_dst_snapshot = dst_snapshots_with_guids[-1].split("\t", 1)
        if p.force_rollback_to_latest_snapshot:
            log.info(p.dry(f"{tid} Rolling back destination to most recent snapshot: %s"), latest_dst_snapshot)
            # rollback just in case the dst dataset was modified since its most recent snapshot
            done_checking = done_checking or _check_zfs_dataset_busy(job, dst, dst_dataset)
            cmd: list[str] = p.split_args(f"{dst.sudo} {p.zfs_program} rollback", latest_dst_snapshot)
            try_ssh_command(job, dst, LOG_DEBUG, is_dry=p.dry_run, print_stdout=True, cmd=cmd, exists=False)
    elif latest_src_snapshot == "":
        log.info(f"{tid} Already-up-to-date: %s", dst_dataset)
        return True

    # find most recent snapshot (or bookmark) that src and dst have in common - we'll start to replicate
    # from there up to the most recent src snapshot. any two snapshots are "common" iff their ZFS GUIDs (i.e.
    # contents) are equal. See https://github.com/openzfs/zfs/commit/305bc4b370b20de81eaf10a1cf724374258b74d1
    def latest_common_snapshot(snapshots_with_guids: list[str], intersect_guids: set[str]) -> tuple[str | None, str]:
        """Returns a true snapshot instead of its bookmark with the same GUID, per the sort order previously used for 'zfs
        list -s ...'."""
        for _line in reversed(snapshots_with_guids):
            guid_, snapshot_ = _line.split("\t", 1)
            if guid_ in intersect_guids:
                return guid_, snapshot_  # can be a snapshot or bookmark
        return None, ""

    latest_common_guid, latest_common_src_snapshot = latest_common_snapshot(
        src_snapshots_with_guids, set(cut(field=1, lines=dst_snapshots_with_guids))
    )
    log.debug("latest_common_src_snapshot: %s", latest_common_src_snapshot)  # is a snapshot or bookmark
    log.log(LOG_TRACE, "latest_dst_snapshot: %s", latest_dst_snapshot)

    if latest_common_src_snapshot and latest_common_guid != latest_dst_guid:
        # found latest common snapshot but dst has an even newer snapshot. rollback dst to that common snapshot.
        assert latest_common_guid
        _, latest_common_dst_snapshot = latest_common_snapshot(dst_snapshots_with_guids, {latest_common_guid})
        if not (p.force_rollback_to_latest_common_snapshot or p.force):
            die(
                f"Conflict: Most recent destination snapshot {latest_dst_snapshot} is more recent than "
                f"most recent common snapshot {latest_common_dst_snapshot}. Rollback destination first, "
                "for example via --force-rollback-to-latest-common-snapshot (or --force) option."
            )
        if p.force_once:
            p.force.value = False
            p.force_rollback_to_latest_common_snapshot.value = False
        log.info(p.dry(f"{tid} Rolling back destination to most recent common snapshot: %s"), latest_common_dst_snapshot)
        done_checking = done_checking or _check_zfs_dataset_busy(job, dst, dst_dataset)
        cmd = p.split_args(
            f"{dst.sudo} {p.zfs_program} rollback -r {p.force_unmount} {p.force_hard}", latest_common_dst_snapshot
        )
        try:
            run_ssh_command(job, dst, LOG_DEBUG, is_dry=p.dry_run, print_stdout=True, cmd=cmd)
        except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
            stderr: str = stderr_to_str(e.stderr) if hasattr(e, "stderr") else ""
            no_sleep: bool = _clear_resumable_recv_state_if_necessary(job, dst_dataset, stderr)
            # op isn't idempotent so retries regather current state from the start of replicate_dataset()
            raise RetryableError("Subprocess failed", no_sleep=no_sleep) from e

    if latest_src_snapshot and latest_src_snapshot == latest_common_src_snapshot:
        log.info(f"{tid} Already up-to-date: %s", dst_dataset)
        return True
    return latest_dst_snapshot, latest_common_src_snapshot, done_checking


def _replicate_dataset_fully(
    job: Job,
    src_dataset: str,
    dst_dataset: str,
    oldest_src_snapshot: str,
    latest_src_snapshot: str,
    latest_dst_snapshot: str,
    dst_snapshots_with_guids: list[str],
    props_cache: dict[tuple[str, str, str], dict[str, str | None]],
    dry_run_no_send: bool,
    done_checking: bool,
    retry_count: int,
    tid: str,
) -> tuple[str, bool, bool, int]:
    """On replication, deletes all dst snapshots and performs a full send of the oldest selected src snapshot, which in turn
    creates a common snapshot."""
    p, log = job.params, job.params.log
    src, dst = p.src, p.dst
    latest_common_src_snapshot: str = ""
    if latest_dst_snapshot:
        if not p.force:
            die(
                f"Conflict: No common snapshot found between {src_dataset} and {dst_dataset} even though "
                "destination has at least one snapshot. Aborting. Consider using --force option to first "
                "delete all existing destination snapshots in order to be able to proceed with replication."
            )
        if p.force_once:
            p.force.value = False
        done_checking = done_checking or _check_zfs_dataset_busy(job, dst, dst_dataset)
        # extract SNAPSHOT_TAG from GUID<TAB>DATASET@SNAPSHOT_TAG
        delete_snapshots(job, dst, dst_dataset, snapshot_tags=cut(2, separator="@", lines=dst_snapshots_with_guids))
        if p.dry_run:
            # As we're in --dryrun (--force) mode this conflict resolution step (see above) wasn't really executed:
            # "no common snapshot was found. delete all dst snapshots". In turn, this would cause the subsequent
            # 'zfs receive -n' to fail with "cannot receive new filesystem stream: destination has snapshots; must
            # destroy them to overwrite it". So we skip the zfs send/receive step and keep on trucking.
            dry_run_no_send = True

    # to start with, fully replicate oldest snapshot, which in turn creates a common snapshot
    if p.no_stream:
        oldest_src_snapshot = latest_src_snapshot
    if oldest_src_snapshot:
        if not job.dst_dataset_exists[dst_dataset]:
            # on destination, create parent filesystem and ancestors if they do not yet exist
            dst_dataset_parent: str = os.path.dirname(dst_dataset)
            if not job.dst_dataset_exists[dst_dataset_parent]:
                if p.dry_run:
                    dry_run_no_send = True
                if dst_dataset_parent:
                    _create_zfs_filesystem(job, dst_dataset_parent)

        recv_resume_token_result: tuple[str | None, list[str], list[str]] = _recv_resume_token(job, dst_dataset, retry_count)
        recv_resume_token, send_resume_opts, recv_resume_opts = recv_resume_token_result
        curr_size: int = _estimate_send_size(job, src, dst_dataset, recv_resume_token, oldest_src_snapshot)
        humansize: str = _format_size(curr_size)
        if recv_resume_token:
            send_opts: list[str] = send_resume_opts  # e.g. ["-t", "1-c740b4779-..."]
        else:
            send_opts = p.curr_zfs_send_program_opts + [oldest_src_snapshot]
        send_cmd: list[str] = p.split_args(f"{src.sudo} {p.zfs_program} send", send_opts)
        recv_opts: list[str] = p.zfs_full_recv_opts.copy() + recv_resume_opts
        recv_opts, set_opts = _add_recv_property_options(job, True, recv_opts, src_dataset, props_cache)
        recv_cmd: list[str] = p.split_args(
            f"{dst.sudo} {p.zfs_program} receive -F", p.dry_run_recv, recv_opts, dst_dataset, allow_all=True
        )
        log.info(p.dry(f"{tid} Full send: %s"), f"{oldest_src_snapshot} --> {dst_dataset} ({humansize.strip()}) ...")
        done_checking = done_checking or _check_zfs_dataset_busy(job, dst, dst_dataset)
        dry_run_no_send = dry_run_no_send or p.dry_run_no_send
        job.maybe_inject_params(error_trigger="full_zfs_send_params")
        humansize = humansize.rjust(RIGHT_JUST * 3 + 2)
        _run_zfs_send_receive(
            job, src_dataset, dst_dataset, send_cmd, recv_cmd, curr_size, humansize, dry_run_no_send, "full_zfs_send"
        )
        latest_common_src_snapshot = oldest_src_snapshot  # we have now created a common snapshot
        if not dry_run_no_send and not p.dry_run:
            job.dst_dataset_exists[dst_dataset] = True
        with job.stats_lock:
            job.num_snapshots_replicated += 1
        _create_zfs_bookmarks(job, src, src_dataset, [oldest_src_snapshot])
        _zfs_set(job, set_opts, dst, dst_dataset)
        dry_run_no_send = dry_run_no_send or p.dry_run
        retry_count = 0

    return latest_common_src_snapshot, dry_run_no_send, done_checking, retry_count


def _replicate_dataset_incrementally(
    job: Job,
    src_dataset: str,
    dst_dataset: str,
    latest_common_src_snapshot: str,
    latest_src_snapshot: str,
    basis_src_snapshots_with_guids: list[str],
    included_src_guids: set[str],
    props_cache: dict[tuple[str, str, str], dict[str, str | None]],
    dry_run_no_send: bool,
    done_checking: bool,
    retry_count: int,
    tid: str,
) -> None:
    """Incrementally replicates all selected snapshots from latest common snapshot until latest src snapshot."""
    p, log = job.params, job.params.log
    src, dst = p.src, p.dst

    def replication_candidates() -> tuple[list[str], list[str]]:
        assert len(basis_src_snapshots_with_guids) > 0
        result_snapshots: list[str] = []
        result_guids: list[str] = []
        last_appended_guid: str = ""
        snapshot_itr: Iterator[str] = reversed(basis_src_snapshots_with_guids)
        while True:
            guid, snapshot = next(snapshot_itr).split("\t", 1)
            if "@" in snapshot:
                result_snapshots.append(snapshot)
                result_guids.append(guid)
                last_appended_guid = guid
            if snapshot == latest_common_src_snapshot:  # latest_common_src_snapshot is a snapshot or bookmark
                if guid != last_appended_guid and "@" not in snapshot:
                    # only appends the src bookmark if it has no snapshot. If the bookmark has a snapshot then that
                    # snapshot has already been appended, per the sort order previously used for 'zfs list -s ...'
                    result_snapshots.append(snapshot)
                    result_guids.append(guid)
                break
        result_snapshots.reverse()
        result_guids.reverse()
        assert len(result_snapshots) > 0
        assert len(result_snapshots) == len(result_guids)
        return result_guids, result_snapshots

    # collect the most recent common snapshot (which may be a bookmark) followed by all src snapshots
    # (that are not a bookmark) that are more recent than that.
    cand_guids, cand_snapshots = replication_candidates()
    if len(cand_snapshots) == 1:
        # latest_src_snapshot is a (true) snapshot that is equal to latest_common_src_snapshot or LESS recent
        # than latest_common_src_snapshot. The latter case can happen if latest_common_src_snapshot is a
        # bookmark whose snapshot has been deleted on src.
        return  # nothing more tbd

    recv_resume_token_result: tuple[str | None, list[str], list[str]] = _recv_resume_token(job, dst_dataset, retry_count)
    recv_resume_token, send_resume_opts, recv_resume_opts = recv_resume_token_result
    recv_opts: list[str] = p.zfs_recv_program_opts.copy() + recv_resume_opts
    recv_opts, set_opts = _add_recv_property_options(job, False, recv_opts, src_dataset, props_cache)
    if p.no_stream:
        # skip intermediate snapshots
        steps_todo: list[tuple[str, str, str, list[str]]] = [
            ("-i", latest_common_src_snapshot, latest_src_snapshot, [latest_src_snapshot])
        ]
    else:
        # include intermediate src snapshots that pass --{include,exclude}-snapshot-* policy, using
        # a series of -i/-I send/receive steps that skip excluded src snapshots.
        steps_todo = _incremental_send_steps_wrapper(
            p, cand_snapshots, cand_guids, included_src_guids, recv_resume_token is not None
        )
    log.log(LOG_TRACE, "steps_todo: %s", list_formatter(steps_todo, "; "))
    estimate_send_sizes: list[int] = _estimate_send_sizes_in_parallel(job, src, dst_dataset, recv_resume_token, steps_todo)
    total_size: int = sum(estimate_send_sizes)
    total_num: int = sum(len(to_snapshots) for incr_flag, from_snap, to_snap, to_snapshots in steps_todo)
    done_size: int = 0
    done_num: int = 0
    for i, (incr_flag, from_snap, to_snap, to_snapshots) in enumerate(steps_todo):
        curr_num_snapshots: int = len(to_snapshots)
        curr_size: int = estimate_send_sizes[i]
        humansize: str = _format_size(total_size) + "/" + _format_size(done_size) + "/" + _format_size(curr_size)
        human_num: str = f"{total_num}/{done_num}/{curr_num_snapshots} snapshots"
        if recv_resume_token:
            send_opts: list[str] = send_resume_opts  # e.g. ["-t", "1-c740b4779-..."]
        else:
            send_opts = p.curr_zfs_send_program_opts + [incr_flag, from_snap, to_snap]
        send_cmd: list[str] = p.split_args(f"{src.sudo} {p.zfs_program} send", send_opts)
        recv_cmd: list[str] = p.split_args(
            f"{dst.sudo} {p.zfs_program} receive", p.dry_run_recv, recv_opts, dst_dataset, allow_all=True
        )
        dense_size: str = p.two_or_more_spaces_regex.sub("", humansize.strip())
        log.info(
            p.dry(f"{tid} Incremental send {incr_flag}: %s"),
            f"{from_snap} .. {to_snap[to_snap.index('@'):]} --> {dst_dataset} ({dense_size}) ({human_num}) ...",
        )
        done_checking = done_checking or _check_zfs_dataset_busy(job, dst, dst_dataset, busy_if_send=False)
        if p.dry_run and not job.dst_dataset_exists[dst_dataset]:
            dry_run_no_send = True
        dry_run_no_send = dry_run_no_send or p.dry_run_no_send
        job.maybe_inject_params(error_trigger="incr_zfs_send_params")
        _run_zfs_send_receive(
            job, src_dataset, dst_dataset, send_cmd, recv_cmd, curr_size, humansize, dry_run_no_send, "incr_zfs_send"
        )
        done_size += curr_size
        done_num += curr_num_snapshots
        recv_resume_token = None
        with job.stats_lock:
            job.num_snapshots_replicated += curr_num_snapshots
        assert p.create_bookmarks
        if p.create_bookmarks == "all":
            _create_zfs_bookmarks(job, src, src_dataset, to_snapshots)
        elif p.create_bookmarks != "none":
            threshold_millis: int = p.xperiods.label_milliseconds("_" + p.create_bookmarks)
            to_snapshots = [snap for snap in to_snapshots if p.xperiods.label_milliseconds(snap) >= threshold_millis]
            if i == len(steps_todo) - 1 and (len(to_snapshots) == 0 or to_snapshots[-1] != to_snap):
                to_snapshots.append(to_snap)  # ensure latest common snapshot is bookmarked
            _create_zfs_bookmarks(job, src, src_dataset, to_snapshots)
    _zfs_set(job, set_opts, dst, dst_dataset)


def _format_size(num_bytes: int) -> str:
    """Formats a byte count for human-readable logs."""
    return human_readable_bytes(num_bytes, separator="").rjust(RIGHT_JUST)


def _prepare_zfs_send_receive(
    job: Job, src_dataset: str, send_cmd: list[str], recv_cmd: list[str], size_estimate_bytes: int, size_estimate_human: str
) -> tuple[str, str, str]:
    """Constructs zfs send/recv pipelines with optional compression and pv."""
    p = job.params
    send_cmd_str: str = shlex.join(send_cmd)
    recv_cmd_str: str = shlex.join(recv_cmd)

    if (
        p.is_program_available("zstd", "src")
        and p.is_program_available("zstd", "dst")
        and p.is_program_available("sh", "src")
        and p.is_program_available("sh", "dst")
    ):
        compress_cmd_: str = _compress_cmd(p, "src", size_estimate_bytes)
        decompress_cmd_: str = _decompress_cmd(p, "dst", size_estimate_bytes)
    else:  # no compression is used if source and destination do not both support compression
        compress_cmd_, decompress_cmd_ = "cat", "cat"

    recordsize: int = abs(job.src_properties[src_dataset].recordsize)
    src_buffer: str = _mbuffer_cmd(p, "src", size_estimate_bytes, recordsize)
    dst_buffer: str = _mbuffer_cmd(p, "dst", size_estimate_bytes, recordsize)
    local_buffer: str = _mbuffer_cmd(p, "local", size_estimate_bytes, recordsize)

    pv_src_cmd: str = ""
    pv_dst_cmd: str = ""
    pv_loc_cmd: str = ""
    if not p.src.ssh_user_host:
        pv_src_cmd = _pv_cmd(job, "local", size_estimate_bytes, size_estimate_human)
    elif not p.dst.ssh_user_host:
        pv_dst_cmd = _pv_cmd(job, "local", size_estimate_bytes, size_estimate_human)
    elif compress_cmd_ == "cat":
        pv_loc_cmd = _pv_cmd(job, "local", size_estimate_bytes, size_estimate_human)  # compression disabled
    else:
        # pull-push mode with compression enabled: reporting "percent complete" isn't straightforward because
        # localhost observes the compressed data instead of the uncompressed data, so we disable the progress bar.
        pv_loc_cmd = _pv_cmd(job, "local", size_estimate_bytes, size_estimate_human, disable_progress_bar=True)

    # assemble pipeline running on source leg
    src_pipe: str = ""
    if job.inject_params.get("inject_src_pipe_fail", False):
        # for testing; initially forward some bytes and then fail
        src_pipe = f"{src_pipe} | dd bs=64 count=1 2>/dev/null && false"
    if job.inject_params.get("inject_src_pipe_garble", False):
        src_pipe = f"{src_pipe} | base64"  # for testing; forward garbled bytes
    if pv_src_cmd and pv_src_cmd != "cat":
        src_pipe = f"{src_pipe} | {pv_src_cmd}"
    if compress_cmd_ != "cat":
        src_pipe = f"{src_pipe} | {compress_cmd_}"
    if src_buffer != "cat":
        src_pipe = f"{src_pipe} | {src_buffer}"
    if src_pipe.startswith(" |"):
        src_pipe = src_pipe[2:]  # strip leading ' |' part
    if job.inject_params.get("inject_src_send_error", False):
        send_cmd_str = f"{send_cmd_str} --injectedGarbageParameter"  # for testing; induce CLI parse error
    if src_pipe:
        src_pipe = f"{send_cmd_str} | {src_pipe}"
        if p.src.ssh_user_host:
            src_pipe = p.shell_program + " -c " + _dquote(src_pipe)
    else:
        src_pipe = send_cmd_str

    # assemble pipeline running on middle leg between source and destination. only enabled for pull-push mode
    local_pipe: str = ""
    if local_buffer != "cat":
        local_pipe = f"{local_buffer}"
    if pv_loc_cmd and pv_loc_cmd != "cat":
        local_pipe = f"{local_pipe} | {pv_loc_cmd}"
        if local_buffer != "cat":
            local_pipe = f"{local_pipe} | {local_buffer}"
    if local_pipe.startswith(" |"):
        local_pipe = local_pipe[2:]  # strip leading ' |' part
    if local_pipe:
        local_pipe = f"| {local_pipe}"

    # assemble pipeline running on destination leg
    dst_pipe: str = ""
    if dst_buffer != "cat":
        dst_pipe = f"{dst_buffer}"
    if decompress_cmd_ != "cat":
        dst_pipe = f"{dst_pipe} | {decompress_cmd_}"
    if pv_dst_cmd and pv_dst_cmd != "cat":
        dst_pipe = f"{dst_pipe} | {pv_dst_cmd}"
    if job.inject_params.get("inject_dst_pipe_fail", False):
        # interrupt zfs receive for testing retry/resume; initially forward some bytes and then stop forwarding
        dst_pipe = f"{dst_pipe} | dd bs=1024 count={INJECT_DST_PIPE_FAIL_KBYTES} 2>/dev/null"
    if job.inject_params.get("inject_dst_pipe_garble", False):
        dst_pipe = f"{dst_pipe} | base64"  # for testing; forward garbled bytes
    if dst_pipe.startswith(" |"):
        dst_pipe = dst_pipe[2:]  # strip leading ' |' part
    if job.inject_params.get("inject_dst_receive_error", False):
        recv_cmd_str = f"{recv_cmd_str} --injectedGarbageParameter"  # for testing; induce CLI parse error
    if dst_pipe:
        dst_pipe = f"{dst_pipe} | {recv_cmd_str}"
        if p.dst.ssh_user_host:
            dst_pipe = p.shell_program + " -c " + _dquote(dst_pipe)
    else:
        dst_pipe = recv_cmd_str

    # If there's no support for shell pipelines, we can't do compression, mbuffering, monitoring and rate-limiting,
    # so we fall back to simple zfs send/receive.
    if not p.is_program_available("sh", "src"):
        src_pipe = send_cmd_str
    if not p.is_program_available("sh", "dst"):
        dst_pipe = recv_cmd_str

    src_pipe = _squote(p.src, src_pipe)
    dst_pipe = _squote(p.dst, dst_pipe)
    return src_pipe, local_pipe, dst_pipe


def _run_zfs_send_receive(
    job: Job,
    src_dataset: str,
    dst_dataset: str,
    send_cmd: list[str],
    recv_cmd: list[str],
    size_estimate_bytes: int,
    size_estimate_human: str,
    dry_run_no_send: bool,
    error_trigger: str | None = None,
) -> None:
    """Executes a zfs send/receive pipeline between source and destination."""
    p, log = job.params, job.params.log
    pipes: tuple[str, str, str] = _prepare_zfs_send_receive(
        job, src_dataset, send_cmd, recv_cmd, size_estimate_bytes, size_estimate_human
    )
    src_pipe, local_pipe, dst_pipe = pipes
    conn_pool_name: str = DEDICATED if p.dedicated_tcp_connection_per_zfs_send else SHARED
    src_conn_pool: ConnectionPool = p.connection_pools["src"].pool(conn_pool_name)
    dst_conn_pool: ConnectionPool = p.connection_pools["dst"].pool(conn_pool_name)
    with src_conn_pool.connection() as src_conn, dst_conn_pool.connection() as dst_conn:
        refresh_ssh_connection_if_necessary(job, p.src, src_conn)
        refresh_ssh_connection_if_necessary(job, p.dst, dst_conn)
        src_ssh_cmd: str = " ".join(src_conn.ssh_cmd_quoted)
        dst_ssh_cmd: str = " ".join(dst_conn.ssh_cmd_quoted)
        cmd: list[str] = [p.shell_program_local, "-c", f"{src_ssh_cmd} {src_pipe} {local_pipe} | {dst_ssh_cmd} {dst_pipe}"]
        msg: str = "Would execute: %s" if dry_run_no_send else "Executing: %s"
        log.debug(msg, cmd[2].lstrip())
        if not dry_run_no_send:
            try:
                maybe_inject_error(job, cmd=cmd, error_trigger=error_trigger)
                sp: Subprocesses = job.subprocesses
                process = sp.subprocess_run(
                    cmd, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, text=True, timeout=timeout(job), check=True
                )
            except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
                no_sleep: bool = False
                if not isinstance(e, UnicodeDecodeError):
                    xprint(log, stderr_to_str(e.stdout), file=sys.stdout)
                    log.warning("%s", stderr_to_str(e.stderr).rstrip())
                if isinstance(e, subprocess.CalledProcessError):
                    no_sleep = _clear_resumable_recv_state_if_necessary(job, dst_dataset, e.stderr)
                # op isn't idempotent so retries regather current state from the start of replicate_dataset()
                raise RetryableError("Subprocess failed", no_sleep=no_sleep) from e
            else:
                xprint(log, process.stdout, file=sys.stdout)
                xprint(log, process.stderr, file=sys.stderr)


def _clear_resumable_recv_state_if_necessary(job: Job, dst_dataset: str, stderr: str) -> bool:
    """Deletes leftover state when resume tokens fail to apply."""

    def clear_resumable_recv_state() -> bool:
        log.warning(p.dry("Aborting an interrupted zfs receive -s, deleting partially received state: %s"), dst_dataset)
        cmd: list[str] = p.split_args(f"{p.dst.sudo} {p.zfs_program} receive -A", dst_dataset)
        try_ssh_command(job, p.dst, LOG_TRACE, is_dry=p.dry_run, print_stdout=True, cmd=cmd)
        log.log(LOG_TRACE, p.dry("Done Aborting an interrupted zfs receive -s: %s"), dst_dataset)
        return True

    p, log = job.params, job.params.log
    # No i18n needed here. OpenZFS ships no translation catalogs, so gettext falls back to English msgids and locale settings
    # have no effect. If translations ever appear, revisit this or inject LC_ALL=C.

    # "cannot resume send: 'wb_src/tmp/src@s1' is no longer the same snapshot used in the initial send"
    # "cannot resume send: 'wb_src/tmp/src@s1' used in the initial send no longer exists"
    # "cannot resume send: incremental source 0xa000000000000000 no longer exists"
    if "cannot resume send" in stderr and (
        "is no longer the same snapshot used in the initial send" in stderr
        or "used in the initial send no longer exists" in stderr
        or re.search(r"incremental source [0-9a-fx]+ no longer exists", stderr)
    ):
        return clear_resumable_recv_state()

    # "cannot receive resume stream: incompatible embedded data stream feature with encrypted receive."
    #     see https://github.com/openzfs/zfs/issues/12480
    # 'cannot receive new filesystem stream: destination xx contains partially-complete state from "zfs receive -s"'
    #     this indicates that --no-resume-recv detects that dst contains a previously interrupted recv -s
    elif "cannot receive" in stderr and (
        "cannot receive resume stream: incompatible embedded data stream feature with encrypted receive" in stderr
        or 'contains partially-complete state from "zfs receive -s"' in stderr
    ):
        return clear_resumable_recv_state()

    elif (  # this signals normal behavior on interrupt of 'zfs receive -s' if running without --no-resume-recv
        "cannot receive new filesystem stream: checksum mismatch or incomplete stream" in stderr
        and "Partially received snapshot is saved" in stderr
    ):
        return True

    # "cannot destroy 'wb_dest/tmp/dst@s1': snapshot has dependent clones ... use '-R' to destroy the following
    # datasets: wb_dest/tmp/dst/%recv" # see https://github.com/openzfs/zfs/issues/10439#issuecomment-642774560
    # This msg indicates a failed 'zfs destroy' via --delete-dst-snapshots. This "clone" is caused by a previously
    # interrupted 'zfs receive -s'. The fix used here is to delete the partially received state of said
    # 'zfs receive -s' via 'zfs receive -A', followed by an automatic retry, which will now succeed to delete the
    # snapshot without user intervention.
    elif (
        "cannot destroy" in stderr
        and "snapshot has dependent clone" in stderr
        and "use '-R' to destroy the following dataset" in stderr
        and f"\n{dst_dataset}/%recv\n" in stderr
    ):
        return clear_resumable_recv_state()

    # Same cause as above, except that this error can occur during 'zfs rollback'
    # Also see https://github.com/openzfs/zfs/blob/master/cmd/zfs/zfs_main.c
    elif (
        "cannot rollback to" in stderr
        and "clones of previous snapshots exist" in stderr
        and "use '-R' to force deletion of the following clones and dependents" in stderr
        and f"\n{dst_dataset}/%recv\n" in stderr
    ):
        return clear_resumable_recv_state()

    return False


def _recv_resume_token(job: Job, dst_dataset: str, retry_count: int) -> tuple[str | None, list[str], list[str]]:
    """Gets recv_resume_token ZFS property from dst_dataset and returns corresponding opts to use for send+recv."""
    p, log = job.params, job.params.log
    if not p.resume_recv:
        return None, [], []
    warning: str | None = None
    if not is_zpool_feature_enabled_or_active(p, p.dst, "feature@extensible_dataset"):
        warning = "not available on destination dataset"
    elif not p.is_program_available(ZFS_VERSION_IS_AT_LEAST_2_1_0, "dst"):
        warning = "unreliable as zfs version is too old"  # e.g. zfs-0.8.3 "internal error: Unknown error 1040"
    if warning:
        log.warning(f"ZFS receive resume feature is {warning}. Falling back to --no-resume-recv: %s", dst_dataset)
        return None, [], []
    recv_resume_token: str | None = None
    send_resume_opts: list[str] = []
    if job.dst_dataset_exists[dst_dataset]:
        cmd: list[str] = p.split_args(f"{p.zfs_program} get -Hp -o value -s none receive_resume_token", dst_dataset)
        recv_resume_token = run_ssh_command(job, p.dst, LOG_TRACE, cmd=cmd).rstrip()
        if recv_resume_token == "-" or not recv_resume_token:  # noqa: S105
            recv_resume_token = None
        else:
            send_resume_opts += ["-n"] if p.dry_run else []
            send_resume_opts += ["-v"] if p.verbose_zfs else []
            send_resume_opts += ["-t", recv_resume_token]
    recv_resume_opts = ["-s"]
    return recv_resume_token, send_resume_opts, recv_resume_opts


def _mbuffer_cmd(p: Params, loc: str, size_estimate_bytes: int, recordsize: int) -> str:
    """If mbuffer command is on the PATH, uses it in the ssh network pipe between 'zfs send' and 'zfs receive' to smooth out
    the rate of data flow and prevent bottlenecks caused by network latency or speed fluctuation."""
    if (
        (p.no_estimate_send_size or size_estimate_bytes >= p.min_pipe_transfer_size)
        and (
            (loc == "src" and (p.src.is_nonlocal or p.dst.is_nonlocal))
            or (loc == "dst" and (p.src.is_nonlocal or p.dst.is_nonlocal))
            or (loc == "local" and p.src.is_nonlocal and p.dst.is_nonlocal)
        )
        and p.is_program_available("mbuffer", loc)
    ):
        recordsize = max(recordsize, 2 * 1024 * 1024)
        return shlex.join([p.mbuffer_program, "-s", str(recordsize)] + p.mbuffer_program_opts)
    else:
        return "cat"


def _compress_cmd(p: Params, loc: str, size_estimate_bytes: int) -> str:
    """If zstd command is on the PATH, uses it in the ssh network pipe between 'zfs send' and 'zfs receive' to reduce network
    bottlenecks by sending compressed data."""
    if (
        (p.no_estimate_send_size or size_estimate_bytes >= p.min_pipe_transfer_size)
        and (p.src.is_nonlocal or p.dst.is_nonlocal)
        and p.is_program_available("zstd", loc)
    ):
        return shlex.join([p.compression_program, "-c"] + p.compression_program_opts)
    else:
        return "cat"


def _decompress_cmd(p: Params, loc: str, size_estimate_bytes: int) -> str:
    """Returns decompression command for network pipe if remote supports it."""
    if (
        (p.no_estimate_send_size or size_estimate_bytes >= p.min_pipe_transfer_size)
        and (p.src.is_nonlocal or p.dst.is_nonlocal)
        and p.is_program_available("zstd", loc)
    ):
        return shlex.join([p.compression_program, "-dc"])
    else:
        return "cat"


WORKER_THREAD_NUMBER_REGEX: Final[re.Pattern[str]] = re.compile(r"ThreadPoolExecutor-\d+_(\d+)")


def _pv_cmd(
    job: Job, loc: str, size_estimate_bytes: int, size_estimate_human: str, disable_progress_bar: bool = False
) -> str:
    """If pv command is on the PATH, monitors the progress of data transfer from 'zfs send' to 'zfs receive'; Progress can be
    viewed via "tail -f $pv_log_file" aka tail -f ~/bzfs-logs/current.pv or similar."""
    p = job.params
    if p.is_program_available("pv", loc):
        size: str = f"--size={size_estimate_bytes}"
        if disable_progress_bar or p.no_estimate_send_size:
            size = ""
        pv_log_file: str = p.log_params.pv_log_file
        thread_name: str = threading.current_thread().name
        if match := WORKER_THREAD_NUMBER_REGEX.fullmatch(thread_name):
            worker = int(match.group(1))
            if worker > 0:
                pv_log_file += PV_FILE_THREAD_SEPARATOR + f"{worker:04}"
        if job.is_first_replication_task.get_and_set(False):
            if job.isatty and not p.quiet:
                job.progress_reporter.start()
            job.replication_start_time_nanos = time.monotonic_ns()
        if job.isatty and not p.quiet:
            with open_nofollow(pv_log_file, mode="a", encoding="utf-8", perm=FILE_PERMISSIONS) as fd:
                fd.write("\n")  # mark start of new stream so ProgressReporter can reliably reset bytes_in_flight
            job.progress_reporter.enqueue_pv_log_file(pv_log_file)
        pv_program_opts: list[str] = [p.pv_program] + p.pv_program_opts
        if job.progress_update_intervals is not None:  # for testing
            pv_program_opts += [f"--interval={job.progress_update_intervals[0]}"]
        pv_program_opts += ["--force", f"--name={size_estimate_human}"]
        pv_program_opts += [size] if size else []
        return f"LC_ALL=C {shlex.join(pv_program_opts)} 2>> {shlex.quote(pv_log_file)}"
    else:
        return "cat"


def _squote(remote: Remote, arg: str) -> str:
    """Quotes an argument only when running remotely over ssh."""
    assert arg is not None
    return shlex.quote(arg) if remote.ssh_user_host else arg


def _dquote(arg: str) -> str:
    """Shell-escapes double quotes and dollar and backticks, then surrounds with double quotes."""
    return '"' + arg.replace('"', '\\"').replace("$", "\\$").replace("`", "\\`") + '"'


def delete_snapshots(job: Job, remote: Remote, dataset: str, snapshot_tags: list[str]) -> None:
    """Deletes snapshots in manageable batches on the specified remote."""
    if len(snapshot_tags) == 0:
        return
    p, log = job.params, job.params.log
    log.info(p.dry(f"Deleting {len(snapshot_tags)} snapshots within %s: %s"), dataset, snapshot_tags)
    # delete snapshots in batches without creating a command line that's too big for the OS to handle
    run_ssh_cmd_batched(
        job,
        remote,
        _delete_snapshot_cmd(p, remote, dataset + "@"),
        snapshot_tags,
        lambda batch: _delete_snapshot(job, remote, dataset, dataset + "@" + ",".join(batch)),
        max_batch_items=job.params.max_snapshots_per_minibatch_on_delete_snaps,
        sep=",",
    )


def _delete_snapshot(job: Job, r: Remote, dataset: str, snapshots_to_delete: str) -> None:
    """Runs zfs destroy for a comma-separated snapshot list."""
    p = job.params
    cmd: list[str] = _delete_snapshot_cmd(p, r, snapshots_to_delete)
    is_dry: bool = False  # False is safe because we're using the 'zfs destroy -n' flag
    try:
        maybe_inject_error(job, cmd=cmd, error_trigger="zfs_delete_snapshot")
        run_ssh_command(job, r, LOG_DEBUG, is_dry=is_dry, print_stdout=True, cmd=cmd)
    except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
        stderr: str = stderr_to_str(e.stderr) if hasattr(e, "stderr") else ""
        no_sleep: bool = _clear_resumable_recv_state_if_necessary(job, dataset, stderr)
        # op isn't idempotent so retries regather current state from the start
        raise RetryableError("Subprocess failed", no_sleep=no_sleep) from e


def _delete_snapshot_cmd(p: Params, r: Remote, snapshots_to_delete: str) -> list[str]:
    """Builds zfs destroy command for given snapshots."""
    return p.split_args(
        f"{r.sudo} {p.zfs_program} destroy", p.force_hard, p.verbose_destroy, p.dry_run_destroy, snapshots_to_delete
    )


def delete_bookmarks(job: Job, remote: Remote, dataset: str, snapshot_tags: list[str]) -> None:
    """Removes bookmarks individually since zfs lacks batch deletion."""
    if len(snapshot_tags) == 0:
        return
    # Unfortunately ZFS has no syntax yet to delete multiple bookmarks in a single CLI invocation
    p, log = job.params, job.params.log
    log.info(
        p.dry(f"Deleting {len(snapshot_tags)} bookmarks within %s: %s"), dataset, dataset + "#" + ",".join(snapshot_tags)
    )
    cmd: list[str] = p.split_args(f"{remote.sudo} {p.zfs_program} destroy")
    run_ssh_cmd_parallel(
        job,
        remote,
        [(cmd, (f"{dataset}#{snapshot_tag}" for snapshot_tag in snapshot_tags))],
        lambda _cmd, batch: try_ssh_command(
            job, remote, LOG_DEBUG, is_dry=p.dry_run, print_stdout=True, cmd=_cmd + batch, exists=False
        ),
        max_batch_items=1,
    )


def delete_datasets(job: Job, remote: Remote, datasets: Iterable[str]) -> None:
    """Deletes the given datasets via zfs destroy -r on the given remote."""
    # Impl is batch optimized to minimize CLI + network roundtrips: only need to run zfs destroy if previously
    # destroyed dataset (within sorted datasets) is not a prefix (aka ancestor) of current dataset
    p, log = job.params, job.params.log
    last_deleted_dataset: str = DONT_SKIP_DATASET
    for dataset in sorted(datasets):
        if is_descendant(dataset, of_root_dataset=last_deleted_dataset):
            continue
        log.info(p.dry("Deleting dataset tree: %s"), f"{dataset} ...")
        cmd: list[str] = p.split_args(
            f"{remote.sudo} {p.zfs_program} destroy -r {p.force_unmount} {p.force_hard} {p.verbose_destroy}",
            p.dry_run_destroy,
            dataset,
        )
        is_dry: bool = False  # False is safe because we're using the 'zfs destroy -n' flag
        run_ssh_command(job, remote, LOG_DEBUG, is_dry=is_dry, print_stdout=True, cmd=cmd)
        last_deleted_dataset = dataset


def _create_zfs_filesystem(job: Job, filesystem: str) -> None:
    """Creates destination filesystem hierarchies without mounting them."""
    # zfs create -p -u $filesystem
    # To ensure the filesystems that we create do not get mounted, we apply a separate 'zfs create -p -u'
    # invocation for each non-existing ancestor. This is because a single 'zfs create -p -u' applies the '-u'
    # part only to the immediate filesystem, rather than to the not-yet existing ancestors.
    p = job.params
    parent: str = ""
    no_mount: str = "-u" if p.is_program_available(ZFS_VERSION_IS_AT_LEAST_2_1_0, "dst") else ""
    for component in filesystem.split("/"):
        parent += component
        if not job.dst_dataset_exists[parent]:
            cmd: list[str] = p.split_args(f"{p.dst.sudo} {p.zfs_program} create -p", no_mount, parent)
            try:
                run_ssh_command(job, p.dst, LOG_DEBUG, is_dry=p.dry_run, print_stdout=True, cmd=cmd)
            except subprocess.CalledProcessError as e:
                # ignore harmless error caused by 'zfs create' without the -u flag, or by dataset already existing
                if not (
                    "filesystem successfully created, but it may only be mounted by root" in e.stderr
                    or "filesystem successfully created, but not mounted" in e.stderr  # SolarisZFS
                    or "dataset already exists" in e.stderr
                    or "filesystem already exists" in e.stderr  # SolarisZFS?
                ):
                    raise
            if not p.dry_run:
                job.dst_dataset_exists[parent] = True
        parent += "/"


def _create_zfs_bookmarks(job: Job, remote: Remote, dataset: str, snapshots: list[str]) -> None:
    """Creates bookmarks for the given snapshots, using the 'zfs bookmark' CLI."""
    # Unfortunately ZFS has no syntax yet to create multiple bookmarks in a single CLI invocation
    p = job.params

    def create_zfs_bookmark(cmd: list[str]) -> None:
        snapshot = cmd[-1]
        assert "@" in snapshot
        bookmark_cmd: list[str] = cmd + [replace_prefix(snapshot, old_prefix=f"{dataset}@", new_prefix=f"{dataset}#")]
        try:
            run_ssh_command(job, remote, LOG_DEBUG, is_dry=p.dry_run, print_stderr=False, cmd=bookmark_cmd)
        except subprocess.CalledProcessError as e:
            # ignore harmless zfs error caused by bookmark with the same name already existing
            if ": bookmark exists" not in e.stderr:
                print(e.stderr, file=sys.stderr, end="")
                raise

    if p.create_bookmarks != "none" and are_bookmarks_enabled(p, remote):
        cmd: list[str] = p.split_args(f"{remote.sudo} {p.zfs_program} bookmark")
        run_ssh_cmd_parallel(
            job, remote, [(cmd, snapshots)], lambda _cmd, batch: create_zfs_bookmark(_cmd + batch), max_batch_items=1
        )


def _estimate_send_size(job: Job, remote: Remote, dst_dataset: str, recv_resume_token: str | None, *items: str) -> int:
    """Estimates num bytes to transfer via 'zfs send -nvP'; Thread-safe."""
    p = job.params
    if p.no_estimate_send_size:
        return 0
    zfs_send_program_opts: list[str] = ["--parsable" if opt == "-P" else opt for opt in p.curr_zfs_send_program_opts]
    zfs_send_program_opts = append_if_absent(zfs_send_program_opts, "-v", "-n", "--parsable")
    if recv_resume_token:
        zfs_send_program_opts = ["-Pnv", "-t", recv_resume_token]
        items = ()
    cmd: list[str] = p.split_args(f"{remote.sudo} {p.zfs_program} send", zfs_send_program_opts, items)
    try:
        lines: str | None = try_ssh_command(job, remote, LOG_TRACE, cmd=cmd)
    except RetryableError as retryable_error:
        assert retryable_error.__cause__ is not None
        if recv_resume_token:
            e = retryable_error.__cause__
            stderr: str = stderr_to_str(e.stderr) if hasattr(e, "stderr") else ""
            retryable_error.no_sleep = _clear_resumable_recv_state_if_necessary(job, dst_dataset, stderr)
        # op isn't idempotent so retries regather current state from the start of replicate_dataset()
        raise
    if lines is None:
        return 0  # src dataset or snapshot has been deleted by third party
    size: str = lines.splitlines()[-1]
    assert size.startswith("size")
    return int(size[size.index("\t") + 1 :])


def _estimate_send_sizes_in_parallel(
    job: Job,
    r: Remote,
    dst_dataset: str,
    recv_resume_token: str | None,
    steps_todo: list[tuple[str, str, str, list[str]]],
) -> list[int]:
    """Estimates num bytes to transfer for multiple send steps; in parallel to reduce latency."""
    p = job.params
    if p.no_estimate_send_size:
        return [0 for _ in steps_todo]

    def iterator_builder(executor: Executor) -> Iterable[Iterable[Future[int]]]:
        resume_token: str | None = recv_resume_token
        return [
            (
                executor.submit(
                    _estimate_send_size, job, r, dst_dataset, resume_token if i == 0 else None, incr_flag, from_snap, to_snap
                )
                for i, (incr_flag, from_snap, to_snap, _to_snapshots) in enumerate(steps_todo)
            )
        ]

    max_workers: int = min(len(steps_todo), job.max_workers[r.location])
    return list(
        parallel_iterator(iterator_builder, max_workers=max_workers, ordered=True, termination_event=job.termination_event)
    )


def _zfs_set(job: Job, properties: list[str], remote: Remote, dataset: str) -> None:
    """Applies the given property key=value pairs via 'zfs set' CLI to the given dataset on the given remote."""
    p = job.params
    if len(properties) == 0:
        return
    # set properties in batches without creating a command line that's too big for the OS to handle
    cmd: list[str] = p.split_args(f"{remote.sudo} {p.zfs_program} set")
    run_ssh_cmd_batched(
        job,
        remote,
        cmd,
        properties,
        lambda batch: run_ssh_command(
            job, remote, LOG_DEBUG, is_dry=p.dry_run, print_stdout=True, cmd=cmd + batch + [dataset]
        ),
        max_batch_items=2**29,
    )


def _zfs_get(
    job: Job,
    remote: Remote,
    dataset: str,
    sources: str,
    output_columns: str,
    propnames: str,
    splitlines: bool,
    props_cache: dict[tuple[str, str, str], dict[str, str | None]],
) -> dict[str, str | None]:
    """Returns the results of 'zfs get' CLI on the given dataset on the given remote."""
    assert dataset
    assert sources
    assert output_columns
    if not propnames:
        return {}
    p = job.params
    cache_key: tuple[str, str, str] = (sources, output_columns, propnames)
    props: dict[str, str | None] | None = props_cache.get(cache_key)
    if props is None:
        cmd: list[str] = p.split_args(f"{p.zfs_program} get -Hp -o {output_columns} -s {sources} {propnames}", dataset)
        lines: str = run_ssh_command(job, remote, LOG_TRACE, cmd=cmd)
        is_name_value_pair: bool = "," in output_columns
        props = {}
        # if not splitlines: omit single trailing newline that was appended by 'zfs get' CLI
        assert splitlines or len(lines) == 0 or lines[-1] == "\n"
        for line in lines.splitlines() if splitlines else [lines[0:-1]]:
            if is_name_value_pair:
                propname, propvalue = line.split("\t", 1)
                props[propname] = propvalue
            else:
                props[line] = None
        props_cache[cache_key] = props
    return props


def _incremental_send_steps_wrapper(
    p: Params, src_snapshots: list[str], src_guids: list[str], included_guids: set[str], is_resume: bool
) -> list[tuple[str, str, str, list[str]]]:
    """Returns incremental send steps, optionally converting -I to -i."""
    force_convert_I_to_i: bool = p.src.use_zfs_delegation and not getenv_bool("no_force_convert_I_to_i", True)  # noqa: N806
    # force_convert_I_to_i == True implies that:
    # If using 'zfs allow' delegation mechanism, force convert 'zfs send -I' to a series of
    # 'zfs send -i' as a workaround for zfs issue https://github.com/openzfs/zfs/issues/16394
    return incremental_send_steps(src_snapshots, src_guids, included_guids, is_resume, force_convert_I_to_i)


def _add_recv_property_options(
    job: Job, full_send: bool, recv_opts: list[str], dataset: str, cache: dict[tuple[str, str, str], dict[str, str | None]]
) -> tuple[list[str], list[str]]:
    """Reads the ZFS properties of the given src dataset; Appends zfs recv -o and -x values to recv_opts according to CLI
    params, and returns properties to explicitly set on the dst dataset after 'zfs receive' completes successfully."""
    p = job.params
    set_opts: list[str] = []
    x_names: list[str] = p.zfs_recv_x_names
    x_names_set: set[str] = set(x_names)
    ox_names: set[str] = p.zfs_recv_ox_names.copy()
    if p.is_program_available(ZFS_VERSION_IS_AT_LEAST_2_2_0, p.dst.location):
        # workaround for https://github.com/openzfs/zfs/commit/b0269cd8ced242e66afc4fa856d62be29bb5a4ff
        # 'zfs recv -x foo' on zfs < 2.2 errors out if the 'foo' property isn't contained in the send stream
        for propname in x_names:
            recv_opts.append("-x")
            recv_opts.append(propname)
        ox_names.update(x_names)  # union
    for config in [p.zfs_recv_o_config, p.zfs_recv_x_config, p.zfs_set_config]:
        if len(config.include_regexes) == 0:
            continue  # this is the default - it's an instant noop
        if (full_send and "full" in config.targets) or (not full_send and "incremental" in config.targets):
            # 'zfs get' uses newline as record separator and tab as separator between output columns. A ZFS user property
            # may contain newline and tab characters (indeed anything). Together, this means that there is no reliable
            # way to determine where a record ends and the next record starts when listing multiple arbitrary records in
            # a single 'zfs get' call. Therefore, here we use a separate 'zfs get' call for each ZFS user property.
            # TODO: perf: on zfs >= 2.3 use json via zfs get -j to safely merge all zfs gets into one 'zfs get' call
            try:
                props_any: dict = _zfs_get(job, p.src, dataset, config.sources, "property", "all", True, cache)
                props_filtered: dict = filter_properties(p, props_any, config.include_regexes, config.exclude_regexes)
                user_propnames: list[str] = [name for name in props_filtered if ":" in name]
                sys_propnames: str = ",".join(name for name in props_filtered if ":" not in name)
                props: dict = _zfs_get(job, p.src, dataset, config.sources, "property,value", sys_propnames, True, cache)
                for propnames in user_propnames:
                    props.update(_zfs_get(job, p.src, dataset, config.sources, "property,value", propnames, False, cache))
            except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
                raise RetryableError("Subprocess failed") from e
            for propname in sorted(props.keys()):
                if config is p.zfs_recv_o_config:
                    if not (propname in ox_names or propname in x_names_set):
                        recv_opts.append("-o")
                        recv_opts.append(f"{propname}={props[propname]}")
                        ox_names.add(propname)
                elif config is p.zfs_recv_x_config:
                    if propname not in ox_names:
                        recv_opts.append("-x")
                        recv_opts.append(propname)
                        ox_names.add(propname)
                else:
                    assert config is p.zfs_set_config
                    set_opts.append(f"{propname}={props[propname]}")
    return recv_opts, set_opts


def _check_zfs_dataset_busy(job: Job, remote: Remote, dataset: str, busy_if_send: bool = True) -> bool:
    """Decline to start a state changing ZFS operation that is, although harmless, likely to collide with other currently
    running processes. Instead, retry the operation later, after some delay. For example, decline to start a 'zfs receive'
    into a destination dataset if another process is already running another 'zfs receive' into the same destination dataset,
    as ZFS would reject any such attempt. However, it's actually fine to run an incremental 'zfs receive' into a dataset in
    parallel with a 'zfs send' out of the very same dataset. This also helps daisy chain use cases where A replicates to B,
    and B replicates to C.

    _check_zfs_dataset_busy() offers no guarantees, it merely proactively avoids likely collisions. In other words, even if
    the process check below passes there is no guarantee that the destination dataset won't be busy by the time we actually
    execute the 'zfs send' operation. In such an event ZFS will reject the operation, we'll detect that, and we'll simply
    retry, after some delay. _check_zfs_dataset_busy() can be disabled via --ps-program=-.

    TLDR: As is common for long-running operations in distributed systems, we use coordination-free optimistic concurrency
    control where the parties simply retry on collision detection (rather than coordinate concurrency via a remote lock
    server).
    """
    p, log = job.params, job.params.log
    if not p.is_program_available("ps", remote.location):
        return True
    cmd: list[str] = p.split_args(f"{p.ps_program} -Ao args")
    procs: list[str] = (try_ssh_command(job, remote, LOG_TRACE, cmd=cmd) or "").splitlines()
    if job.inject_params.get("is_zfs_dataset_busy", False):
        procs += ["sudo -n zfs receive -u -o foo:bar=/baz " + dataset]  # for unit testing only
    if not _is_zfs_dataset_busy(procs, dataset, busy_if_send=busy_if_send):
        return True
    op: str = "zfs {receive" + ("|send" if busy_if_send else "") + "} operation"
    try:
        die(f"Cannot continue now: Destination is already busy with {op} from another process: {dataset}")
    except SystemExit as e:
        log.warning("%s", e)
        raise RetryableError("dst currently busy with zfs mutation op") from e


ZFS_DATASET_BUSY_PREFIX: Final[str] = r"(([^ ]*?/)?(sudo|doas)( +-n)? +)?([^ ]*?/)?zfs (receive|recv"
ZFS_DATASET_BUSY_IF_MODS: Final[re.Pattern[str]] = re.compile((ZFS_DATASET_BUSY_PREFIX + ") .*").replace("(", "(?:"))
ZFS_DATASET_BUSY_IF_SEND: Final[re.Pattern[str]] = re.compile((ZFS_DATASET_BUSY_PREFIX + "|send) .*").replace("(", "(?:"))


def _is_zfs_dataset_busy(procs: list[str], dataset: str, busy_if_send: bool) -> bool:
    """Checks if any process list entry indicates ZFS activity on dataset."""
    regex: re.Pattern[str] = ZFS_DATASET_BUSY_IF_SEND if busy_if_send else ZFS_DATASET_BUSY_IF_MODS
    suffix: str = " " + dataset
    infix: str = " " + dataset + "@"
    return any((proc.endswith(suffix) or infix in proc) and regex.fullmatch(proc) for proc in procs)



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Network connection management is in refresh_ssh_connection_if_necessary() and class ConnectionPool; They reuse multiplexed
ssh connections for low latency."""

from __future__ import (
    annotations,
)
import contextlib
import copy
import logging
import shlex
import subprocess
import threading
import time
from collections import (
    Counter,
)
from collections.abc import (
    Iterator,
)
from dataclasses import (
    dataclass,
)
from subprocess import (
    DEVNULL,
    PIPE,
    CalledProcessError,
    CompletedProcess,
)
from typing import (
    TYPE_CHECKING,
    Final,
)

from bzfs_main.connection_lease import (
    ConnectionLease,
    ConnectionLeaseManager,
)
from bzfs_main.retry import (
    RetryableError,
)
from bzfs_main.utils import (
    LOG_TRACE,
    PROG_NAME,
    SmallPriorityQueue,
    Subprocesses,
    die,
    list_formatter,
    stderr_to_str,
    xprint,
)

if TYPE_CHECKING:  # pragma: no cover - for type hints only
    from bzfs_main.bzfs import (
        Job,
    )
    from bzfs_main.configuration import (
        Params,
        Remote,
    )

# constants:
SHARED: Final[str] = "shared"
DEDICATED: Final[str] = "dedicated"


def run_ssh_command(
    job: Job,
    remote: Remote,
    level: int = -1,
    is_dry: bool = False,
    check: bool = True,
    print_stdout: bool = False,
    print_stderr: bool = True,
    cmd: list[str] | None = None,
) -> str:
    """Runs the given CLI cmd via ssh on the given remote, and returns stdout.

    The full command is the concatenation of both the command to run on the localhost in order to talk to the remote host
    ($remote.local_ssh_command()) and the command to run on the given remote host ($cmd).

    Note: When executing on a remote host (remote.ssh_user_host is set), cmd arguments are pre-quoted with shlex.quote to
    safely traverse the ssh "remote shell" boundary, as ssh concatenates argv into a single remote shell string. In local
    mode (no remote.ssh_user_host) argv is executed directly without an intermediate shell.
    """
    level = level if level >= 0 else logging.INFO
    assert cmd is not None and isinstance(cmd, list) and len(cmd) > 0
    p, log = job.params, job.params.log
    quoted_cmd: list[str] = [shlex.quote(arg) for arg in cmd]
    conn_pool: ConnectionPool = p.connection_pools[remote.location].pool(SHARED)
    with conn_pool.connection() as conn:
        ssh_cmd: list[str] = conn.ssh_cmd
        if remote.ssh_user_host:
            refresh_ssh_connection_if_necessary(job, remote, conn)
            cmd = quoted_cmd
        msg: str = "Would execute: %s" if is_dry else "Executing: %s"
        log.log(level, msg, list_formatter(conn.ssh_cmd_quoted + quoted_cmd, lstrip=True))
        if is_dry:
            return ""
        try:
            sp: Subprocesses = job.subprocesses
            process: CompletedProcess[str] = sp.subprocess_run(
                ssh_cmd + cmd, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, text=True, timeout=timeout(job), check=check
            )
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired, UnicodeDecodeError) as e:
            if not isinstance(e, UnicodeDecodeError):
                xprint(log, stderr_to_str(e.stdout), run=print_stdout, end="")
                xprint(log, stderr_to_str(e.stderr), run=print_stderr, end="")
            raise
        else:
            xprint(log, process.stdout, run=print_stdout, end="")
            xprint(log, process.stderr, run=print_stderr, end="")
            return process.stdout


def try_ssh_command(
    job: Job,
    remote: Remote,
    level: int,
    is_dry: bool = False,
    print_stdout: bool = False,
    cmd: list[str] | None = None,
    exists: bool = True,
    error_trigger: str | None = None,
) -> str | None:
    """Convenience method that helps retry/react to a dataset or pool that potentially doesn't exist anymore."""
    assert cmd is not None and isinstance(cmd, list) and len(cmd) > 0
    log = job.params.log
    try:
        maybe_inject_error(job, cmd=cmd, error_trigger=error_trigger)
        return run_ssh_command(job, remote, level=level, is_dry=is_dry, print_stdout=print_stdout, cmd=cmd)
    except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
        if not isinstance(e, UnicodeDecodeError):
            stderr: str = stderr_to_str(e.stderr)
            if exists and (
                ": dataset does not exist" in stderr
                or ": filesystem does not exist" in stderr  # solaris 11.4.0
                or ": no such pool" in stderr
            ):
                return None
            log.warning("%s", stderr.rstrip())
        raise RetryableError("Subprocess failed") from e


def refresh_ssh_connection_if_necessary(job: Job, remote: Remote, conn: Connection) -> None:
    """Maintain or create an ssh master connection for low latency reuse."""
    p, log = job.params, job.params.log
    if not remote.ssh_user_host:
        return  # we're in local mode; no ssh required
    if not p.is_program_available("ssh", "local"):
        die(f"{p.ssh_program} CLI is not available to talk to remote host. Install {p.ssh_program} first!")
    if not remote.reuse_ssh_connection:
        return
    # Performance: reuse ssh connection for low latency startup of frequent ssh invocations via the 'ssh -S' and
    # 'ssh -S -M -oControlPersist=90s' options. See https://en.wikibooks.org/wiki/OpenSSH/Cookbook/Multiplexing
    # and https://chessman7.substack.com/p/how-ssh-multiplexing-reuses-master
    control_persist_limit_nanos: int = (remote.ssh_control_persist_secs - job.control_persist_margin_secs) * 1_000_000_000
    with conn.lock:
        if time.monotonic_ns() < conn.last_refresh_time + control_persist_limit_nanos:
            return  # ssh master is alive, reuse its TCP connection (this is the common case and the ultra-fast path)
        ssh_cmd: list[str] = conn.ssh_cmd
        ssh_socket_cmd: list[str] = ssh_cmd[0:-1]  # omit trailing ssh_user_host
        ssh_socket_cmd += ["-O", "check", remote.ssh_user_host]
        # extend lifetime of ssh master by $ssh_control_persist_secs via 'ssh -O check' if master is still running.
        # 'ssh -S /path/to/socket -O check' doesn't talk over the network, hence is still a low latency fast path.
        sp: Subprocesses = job.subprocesses
        if sp.subprocess_run(ssh_socket_cmd, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, timeout=timeout(job)).returncode == 0:
            log.log(LOG_TRACE, "ssh connection is alive: %s", list_formatter(ssh_socket_cmd))
        else:  # ssh master is not alive; start a new master:
            log.log(LOG_TRACE, "ssh connection is not yet alive: %s", list_formatter(ssh_socket_cmd))
            ssh_control_persist_secs: int = remote.ssh_control_persist_secs
            if "-v" in remote.ssh_extra_opts:
                # Unfortunately, with `ssh -v` (debug mode), the ssh master won't background; instead it stays in the
                # foreground and blocks until the ControlPersist timer expires (90 secs). To make progress earlier we ...
                ssh_control_persist_secs = min(1, ssh_control_persist_secs)  # tell ssh to block as briefly as possible (1s)
            ssh_socket_cmd = ssh_cmd[0:-1]  # omit trailing ssh_user_host
            ssh_socket_cmd += ["-M", f"-oControlPersist={ssh_control_persist_secs}s", remote.ssh_user_host, "exit"]
            log.log(LOG_TRACE, "Executing: %s", list_formatter(ssh_socket_cmd))
            try:
                sp.subprocess_run(ssh_socket_cmd, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, check=True, timeout=timeout(job))
            except subprocess.CalledProcessError as e:
                log.error("%s", stderr_to_str(e.stderr).rstrip())
                raise RetryableError(
                    f"Cannot ssh into remote host via '{' '.join(ssh_socket_cmd)}'. Fix ssh configuration "
                    f"first, considering diagnostic log file output from running {PROG_NAME} with -v -v -v."
                ) from e
        conn.last_refresh_time = time.monotonic_ns()
        if conn.connection_lease is not None:
            conn.connection_lease.set_socket_mtime_to_now()


def timeout(job: Job) -> float | None:
    """Raises TimeoutExpired if necessary, else returns the number of seconds left until timeout is to occur."""
    timeout_nanos: int | None = job.timeout_nanos
    if timeout_nanos is None:
        return None  # never raise a timeout
    delta_nanos: int = timeout_nanos - time.monotonic_ns()
    if delta_nanos <= 0:
        assert job.params.timeout_nanos is not None
        raise subprocess.TimeoutExpired(PROG_NAME + "_timeout", timeout=job.params.timeout_nanos / 1_000_000_000)
    return delta_nanos / 1_000_000_000  # seconds


def maybe_inject_error(job: Job, cmd: list[str], error_trigger: str | None = None) -> None:
    """For testing only; for unit tests to simulate errors during replication and test correct handling of them."""
    if error_trigger:
        counter = job.error_injection_triggers.get("before")
        if counter and decrement_injection_counter(job, counter, error_trigger):
            try:
                raise CalledProcessError(returncode=1, cmd=" ".join(cmd), stderr=error_trigger + ":dataset is busy")
            except subprocess.CalledProcessError as e:
                if error_trigger.startswith("retryable_"):
                    raise RetryableError("Subprocess failed") from e
                else:
                    raise


def decrement_injection_counter(job: Job, counter: Counter[str], trigger: str) -> bool:
    """For testing only."""
    with job.injection_lock:
        if counter[trigger] <= 0:
            return False
        counter[trigger] -= 1
        return True


#############################################################################
@dataclass(order=True, repr=False)
class Connection:
    """Represents the ability to multiplex N=capacity concurrent SSH sessions over the same TCP connection."""

    _free: int  # sort order evens out the number of concurrent sessions among the TCP connections
    _last_modified: int  # LIFO: tiebreaker favors latest returned conn as that's most alive and hot; also ensures no dupes

    def __init__(
        self,
        remote: Remote,
        max_concurrent_ssh_sessions_per_tcp_connection: int,
        cid: int,
        lease: ConnectionLease | None = None,
    ) -> None:
        assert max_concurrent_ssh_sessions_per_tcp_connection > 0
        self._capacity: Final[int] = max_concurrent_ssh_sessions_per_tcp_connection
        self._free: int = max_concurrent_ssh_sessions_per_tcp_connection
        self._last_modified: int = 0  # monotonically increasing
        self._cid: Final[int] = cid
        self.last_refresh_time: int = 0
        self.lock: Final[threading.Lock] = threading.Lock()
        self._reuse_ssh_connection: Final[bool] = remote.reuse_ssh_connection
        self.connection_lease: Final[ConnectionLease | None] = lease
        self.ssh_cmd: Final[list[str]] = remote.local_ssh_command(
            None if self.connection_lease is None else self.connection_lease.socket_path
        )
        self.ssh_cmd_quoted: Final[list[str]] = [shlex.quote(item) for item in self.ssh_cmd]

    def __repr__(self) -> str:
        return str({"free": self._free, "cid": self._cid})

    def increment_free(self, value: int) -> None:
        """Adjusts the count of available SSH slots."""
        self._free += value
        assert self._free >= 0
        assert self._free <= self._capacity

    def is_full(self) -> bool:
        """Returns True if no more SSH sessions may be opened over this TCP connection."""
        return self._free <= 0

    def update_last_modified(self, last_modified: int) -> None:
        """Records when the connection was last used."""
        self._last_modified = last_modified

    def shutdown(self, msg_prefix: str, p: Params) -> None:
        """Closes the underlying SSH master connection and releases the corresponding connection lease."""
        ssh_cmd: list[str] = self.ssh_cmd
        if ssh_cmd and self._reuse_ssh_connection:
            if self.connection_lease is None:
                ssh_sock_cmd: list[str] = ssh_cmd[0:-1] + ["-O", "exit", ssh_cmd[-1]]
                p.log.log(LOG_TRACE, f"Executing {msg_prefix}: %s", shlex.join(ssh_sock_cmd))
                try:
                    proc: CompletedProcess = subprocess.run(ssh_sock_cmd, stdin=DEVNULL, stderr=PIPE, text=True, timeout=0.1)
                except subprocess.TimeoutExpired as e:  # harmless as master auto-exits after ssh_control_persist_secs anyway
                    p.log.log(LOG_TRACE, "Harmless ssh master connection shutdown timeout: %s", e)
                else:
                    if proc.returncode != 0:  # harmless for the same reason
                        p.log.log(LOG_TRACE, "Harmless ssh master connection shutdown issue: %s", proc.stderr.rstrip())
            else:
                self.connection_lease.release()


#############################################################################
class ConnectionPool:
    """Fetch a TCP connection for use in an SSH session, use it, finally return it back to the pool for future reuse."""

    def __init__(self, remote: Remote, max_concurrent_ssh_sessions_per_tcp_connection: int, connpool_name: str) -> None:
        assert max_concurrent_ssh_sessions_per_tcp_connection > 0
        self._remote: Final[Remote] = copy.copy(remote)  # shallow copy for immutability (Remote is mutable)
        self._capacity: Final[int] = max_concurrent_ssh_sessions_per_tcp_connection
        self._connpool_name: Final[str] = connpool_name
        self._priority_queue: Final[SmallPriorityQueue[Connection]] = SmallPriorityQueue(
            reverse=True  # sorted by #free slots and last_modified
        )
        self._last_modified: int = 0  # monotonically increasing sequence number
        self._cid: int = 0  # monotonically increasing connection number
        self._lock: Final[threading.Lock] = threading.Lock()
        lease_mgr: ConnectionLeaseManager | None = None
        if self._remote.ssh_user_host and self._remote.reuse_ssh_connection and not self._remote.ssh_exit_on_shutdown:
            lease_mgr = ConnectionLeaseManager(
                root_dir=self._remote.ssh_socket_dir,
                namespace=f"{self._remote.location}#{self._remote.cache_namespace()}#{self._connpool_name}",
                ssh_control_persist_secs=max(90 * 60, 2 * self._remote.ssh_control_persist_secs + 2),
                log=self._remote.params.log,
            )
        self._lease_mgr: Final[ConnectionLeaseManager | None] = lease_mgr

    @contextlib.contextmanager
    def connection(self) -> Iterator[Connection]:
        """Context manager that yields a connection from the pool and automatically returns it on __exit__."""
        conn: Connection = self.get_connection()
        try:
            yield conn
        finally:
            self.return_connection(conn)

    def get_connection(self) -> Connection:
        """Any Connection object returned on get_connection() also remains intentionally contained in the priority queue
        while it is "checked out", and that identical Connection object is later, on return_connection(), temporarily removed
        from the priority queue, updated with an incremented "free" slot count and then immediately reinserted into the
        priority queue.

        In effect, any Connection object remains intentionally contained in the priority queue at all times. This design
        keeps ordering/fairness accurate while avoiding duplicate Connection instances.
        """
        with self._lock:
            conn = self._priority_queue.pop() if len(self._priority_queue) > 0 else None
            if conn is None or conn.is_full():
                if conn is not None:
                    self._priority_queue.push(conn)
                lease: ConnectionLease | None = None if self._lease_mgr is None else self._lease_mgr.acquire()
                conn = Connection(self._remote, self._capacity, self._cid, lease=lease)  # add a new connection
                self._last_modified += 1
                conn.update_last_modified(self._last_modified)  # LIFO tiebreaker favors latest conn as that's most alive
                self._cid += 1
            conn.increment_free(-1)
            self._priority_queue.push(conn)
            return conn

    def return_connection(self, conn: Connection) -> None:
        """Returns the given connection to the pool and updates its priority."""
        assert conn is not None
        with self._lock:
            # update priority = remove conn from queue, increment priority, finally reinsert updated conn into queue
            if self._priority_queue.remove(conn):  # conn is not contained only if ConnectionPool.shutdown() was called
                conn.increment_free(1)
                self._last_modified += 1
                conn.update_last_modified(self._last_modified)  # LIFO tiebreaker favors latest conn as that's most alive
                self._priority_queue.push(conn)

    def shutdown(self, msg_prefix: str) -> None:
        """Closes all SSH connections managed by this pool."""
        with self._lock:
            try:
                if self._remote.reuse_ssh_connection:
                    for conn in self._priority_queue:
                        conn.shutdown(msg_prefix, self._remote.params)
            finally:
                self._priority_queue.clear()

    def __repr__(self) -> str:
        with self._lock:
            queue = self._priority_queue
            return str({"capacity": self._capacity, "queue_len": len(queue), "cid": self._cid, "queue": queue})


#############################################################################
class ConnectionPools:
    """A bunch of named connection pools with various multiplexing capacities."""

    def __init__(self, remote: Remote, capacities: dict[str, int]) -> None:
        """Creates one connection pool per name with the given capacities."""
        self._pools: Final[dict[str, ConnectionPool]] = {
            name: ConnectionPool(remote, capacity, name) for name, capacity in capacities.items()
        }

    def __repr__(self) -> str:
        return str(self._pools)

    def pool(self, name: str) -> ConnectionPool:
        """Returns the pool associated with the given name."""
        return self._pools[name]

    def shutdown(self, msg_prefix: str) -> None:
        """Shuts down every contained pool."""
        for name, pool in self._pools.items():
            pool.shutdown(msg_prefix + "/" + name)



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper functions for running ZFS CLI commands, parsing and validating ZFS data."""

from __future__ import (
    annotations,
)
import re
import subprocess
from collections.abc import (
    Mapping,
    Sequence,
)
from typing import (
    cast,
)

sudo_cmd: list[str] = []


def set_sudo_cmd(sudo: list[str]) -> None:
    global sudo_cmd
    sudo_cmd = sudo


def destroy_pool(pool_name: str, force: bool = False) -> None:
    force_flag = ["-f"] if force else []
    run_cmd(sudo_cmd + ["zpool", "destroy"] + force_flag + [pool_name])


def destroy(name: str, recursive: bool = False, force: bool = False) -> None:
    cmd = sudo_cmd + ["zfs", "destroy"]
    if recursive:
        cmd.append("-r")
    if force:
        cmd.append("-f")
    cmd.append(name)
    run_cmd(cmd)


def destroy_snapshots(dataset: str, snapshots: Sequence[str] = ()) -> None:
    if len(snapshots) == 0:
        return
    cmd = sudo_cmd + ["zfs", "destroy"]
    cmd += [dataset + "@" + ",".join([snap[snap.find("@") + 1 :] for snap in snapshots])]
    run_cmd(cmd)


def create_volume(
    dataset: str,
    path: str | None = None,
    mk_parents: bool = True,
    size: int | str | None = None,
    props: list[str] | None = None,
    blocksize: int | None = None,
    sparse: bool = False,
) -> str:
    if props is None:
        props = []
    path = "" if path is None else "/" + path
    dataset = f"{dataset}{path}"
    cmd = sudo_cmd + ["zfs", "create"]
    if mk_parents:
        cmd.append("-p")
    if sparse:
        cmd.append("-s")
    if blocksize:
        cmd.append("-b")
        cmd.append(str(blocksize))
    if props:
        cmd += props
    cmd.append("-V")
    cmd.append(str(size))
    cmd.append(dataset)
    run_cmd(cmd)
    return dataset


def create_filesystem_simple(
    dataset: str,
    path: str | None = None,
    mk_parents: bool = True,
    no_mount: bool = True,
    props: list[str] | None = None,
) -> str:
    if props is None:
        props = []
    path = "" if path is None else "/" + path
    dataset = f"{dataset}{path}"
    cmd = sudo_cmd + ["zfs", "create"]
    if mk_parents:
        cmd += ["-p"]
    if no_mount:
        cmd += ["-u"]
    if props:
        cmd += props
    cmd.append(dataset)
    run_cmd(cmd)
    return dataset


zfs_version_is_at_least_2_1_0 = None


def create_filesystem(
    dataset: str,
    path: str | None = None,
    no_mount: bool = True,
    props: list[str] | None = None,
) -> str:
    """Implies mk_parents=True if no_mount=True:

    To ensure the datasets that we create do not get mounted, we apply a separate 'zfs create -p -u' invocation for each non-
    existing ancestor. This is because a single 'zfs create -p -u' applies the '-u' part only to the immediate dataset,
    rather than to the not-yet existing ancestors. If the zfs version doesn't support the 'zfs create -u' flag then we
    manually unmount immediately after each dataset creation.
    """
    if props is None:
        props = []
    mk_parents = True
    path = "" if path is None else "/" + path
    dataset = f"{dataset}{path}"
    parent = ""
    splits = dataset.split("/")
    for i, component in enumerate(splits):
        parent += component
        cmd = sudo_cmd + ["zfs", "create"]
        if mk_parents:
            cmd.append("-p")
        global zfs_version_is_at_least_2_1_0
        if no_mount:
            if zfs_version_is_at_least_2_1_0 is None:
                version = zfs_version()
                if version is None:
                    version = "0.0.0"
                zfs_version_is_at_least_2_1_0 = is_version_at_least(version, "2.1.0")
            if zfs_version_is_at_least_2_1_0:
                cmd.append("-u")
            else:
                already_exists = dataset_exists(parent)

        if props and i == len(splits) - 1:  # apply props to leaf dataset only, just like 'zfs create -p dataset'
            cmd += props
        cmd.append(parent)
        run_cmd(cmd)

        if no_mount and (not zfs_version_is_at_least_2_1_0) and not already_exists:
            # zfs < 2.1.0 does not know the zfs create -u flag, so we manually unmount after dataset creation
            run_cmd(sudo_cmd + ["zfs", "unmount", parent])

        parent += "/"
    return dataset


def datasets(dataset: str) -> list[str]:
    return cast(list[str], zfs_list([dataset], types=["filesystem", "volume"], max_depth=1))[1:]


def take_snapshot(
    dataset: str,
    snapshot_tag: str,
    recursive: bool = False,
    props: list[str] | None = None,
) -> str:
    if props is None:
        props = []
    snapshot = dataset + "@" + snapshot_tag
    cmd = sudo_cmd + ["zfs", "snapshot"]
    if recursive:
        cmd.append("-r")
    if props:
        cmd += props
    cmd.append(snapshot)
    run_cmd(cmd)
    return snapshot


def snapshots(dataset: str, max_depth: int | None = 1) -> list[str]:
    return cast(list[str], zfs_list([dataset], types=["snapshot"], max_depth=max_depth))


def create_bookmark(dataset: str, snapshot_tag: str, bookmark_tag: str) -> str:
    snapshot = dataset + "@" + snapshot_tag
    bookmark = dataset + "#" + bookmark_tag
    run_cmd(sudo_cmd + ["zfs", "bookmark", snapshot, bookmark])
    return bookmark


def bookmarks(dataset: str, max_depth: int = 1) -> list[str]:
    return cast(list[str], zfs_list([dataset], types=["bookmark"], max_depth=max_depth))


def snapshot_name(snapshot: str) -> str:
    return snapshot[snapshot.find("@") + 1 :]


def bookmark_name(bookmark: str) -> str:
    return bookmark[bookmark.find("#") + 1 :]


def dataset_property(dataset: str, prop: str) -> str:
    return cast(str, zfs_list([dataset], props=[prop], types=["filesystem", "volume"], max_depth=0, splitlines=False))
    # return zfs_get(
    #     [dataset], props=[prop], types=["filesystem", "volume"], max_depth=0, fields=["value"], splitlines=False
    # )


def snapshot_property(snapshot: str, prop: str) -> str:
    return cast(str, zfs_list([snapshot], props=[prop], types=["snapshot"], max_depth=0, splitlines=False))


def zfs_list(
    names: list[str] | None = None,
    props: list[str] | None = None,
    types: list[str] | None = None,
    max_depth: int | None = None,
    parsable: bool = True,
    sort_props: list[str] | None = None,
    splitlines: bool = True,
) -> list[str] | str:
    cmd = ["zfs", "list"]
    if names is None:
        names = []
    if props is None:
        props = ["name"]
    if types is None:
        types = []
    if sort_props is None:
        sort_props = []
    if max_depth is None:
        cmd.append("-r")
    else:
        cmd.append("-d")
        cmd.append(str(max_depth))

    cmd.append("-H")
    if parsable:
        cmd.append("-p")

    if props:
        cmd.append("-o")
        cmd.append(",".join(props))

    if sort_props:
        cmd += sort_props

    if types:
        cmd.append("-t")
        cmd.append(",".join(types))

    if names:
        cmd += names

    return run_cmd(cmd, splitlines=splitlines)


def zfs_get(
    names: list[str] | None = None,
    props: list[str] | None = None,
    types: list[str] | None = None,
    max_depth: int | None = None,
    parsable: bool = True,
    fields: list[str] | None = None,
    sources: list[str] | None = None,
    splitlines: bool = True,
) -> list[str] | str:
    cmd = ["zfs", "get"]
    if names is None:
        names = []
    if props is None:
        props = ["all"]
    if types is None:
        types = []
    if fields is None:
        fields = []
    if sources is None:
        sources = []
    if max_depth is None:
        cmd.append("-r")
    else:
        cmd.append("-d")
        cmd.append(str(max_depth))

    cmd.append("-H")
    if parsable:
        cmd.append("-p")

    if fields:  # defaults to name,property,value,source
        cmd.append("-o")
        cmd.append(",".join(fields))

    if sources:  # 'local', 'default', 'inherited', 'temporary', 'received', 'none'. default is all sources
        cmd.append("-s")
        cmd.append(",".join(sources))

    if types:  # filesystem, snapshot, volume, bookmark, or all. default is all
        cmd.append("-t")
        cmd.append(",".join(types))

    assert len(props) > 0
    cmd.append(",".join(props))

    if names:
        cmd += names

    return run_cmd(cmd, splitlines=splitlines)


def zfs_set(names: list[str] | None = None, properties: Mapping[str, str] | None = None) -> None:
    if names is None:
        names = []
    if properties is None:
        properties = {}

    def run_zfs_set(props: list[str]) -> None:
        cmd = sudo_cmd + ["zfs", "set"]
        for prop in props:
            cmd.append(prop)
        if names:
            cmd += names
        run_cmd(cmd)

    run_zfs_set([f"{name}={value}" for name, value in properties.items()])


def zfs_inherit(
    names: list[str] | None = None,
    propname: str | None = None,
    recursive: bool = False,
    revert: bool = False,
) -> None:
    if names is None:
        names = []
    assert propname
    cmd = sudo_cmd + ["zfs", "inherit"]
    if recursive:
        cmd.append("-r")
    if revert:
        cmd.append("-S")
    cmd.append(propname)
    if names:
        cmd += names

    run_cmd(cmd)


def dataset_exists(dataset: str) -> bool:
    try:
        build(dataset)
        return True
    except subprocess.CalledProcessError:
        return False


def build(name: str, check: bool = True) -> str:
    if check:
        if "@" in name:
            types = ["snapshot"]
        elif "#" in name:
            types = ["bookmark"]
        else:
            types = ["filesystem", "volume"]

        if len(zfs_list([name], types=types, max_depth=0)) == 0:
            raise ValueError("Cannot zfs_list: " + name)

    return name


def zfs_version() -> str:
    """Example zfs-2.1.5~rc5-ubuntu3 -> 2.1.5."""
    try:
        # on Linux, 'zfs --version' returns with zero status and prints the correct info
        # on FreeBSD, 'zfs --version' always prints the same (correct) info as Linux, but nonetheless sometimes
        # returns with non-zero status (sometimes = if the zfs kernel module is not loaded)
        lines = subprocess.run(["zfs", "--version"], capture_output=True, text=True, check=True).stdout
        assert lines
    except subprocess.CalledProcessError as e:
        if "unrecognized command '--version'" in e.stderr and "run: zfs help" in e.stderr:
            raise
        elif not e.stdout.startswith("zfs"):
            raise
        else:
            lines = e.stdout  # FreeBSD if the zfs kernel module is not loaded
            assert lines

    line = lines.splitlines()[0]
    version = line.split("-")[1].strip()
    match = re.fullmatch(r"(\d+\.\d+\.\d+).*", version)
    if match:
        return match.group(1)
    else:
        raise ValueError("Unparsable zfs version string: " + version)


def is_version_at_least(version_str: str, min_version_str: str) -> bool:
    """Check if the version string is at least the minimum version string."""
    return tuple(map(int, version_str.split("."))) >= tuple(map(int, min_version_str.split(".")))


def run_cmd(cmd: Sequence[str], splitlines: bool = True) -> list[str] | str:
    stdout = subprocess.run(cmd, stdout=subprocess.PIPE, text=True, check=True).stdout
    assert stdout is not None
    return stdout.splitlines() if splitlines else stdout[0:-1]  # omit trailing newline char



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility that snaps datetimes to calendar periods.

Anchors specify offsets within yearly, monthly and smaller cycles. These values are used by
``round_datetime_up_to_duration_multiple`` to snap datetimes to the next boundary. Keeping anchors in a dataclass simplifies
argument handling and makes the rounding logic reusable.
"""

from __future__ import (
    annotations,
)
import argparse
import calendar
import dataclasses
from dataclasses import (
    dataclass,
    field,
)
from datetime import (
    datetime,
    timedelta,
)
from typing import (
    Final,
)

# constants:
METADATA_MONTH: Final = {"min": 1, "max": 12, "help": "The month within a year"}
METADATA_WEEKDAY: Final = {"min": 0, "max": 6, "help": "The weekday within a week: 0=Sunday, 1=Monday, ..., 6=Saturday"}
METADATA_DAY: Final = {"min": 1, "max": 31, "help": "The day within a month"}
METADATA_HOUR: Final = {"min": 0, "max": 23, "help": "The hour within a day"}
METADATA_MINUTE: Final = {"min": 0, "max": 59, "help": "The minute within an hour"}
METADATA_SECOND: Final = {"min": 0, "max": 59, "help": "The second within a minute"}
METADATA_MILLISECOND: Final = {"min": 0, "max": 999, "help": "The millisecond within a second"}
METADATA_MICROSECOND: Final = {"min": 0, "max": 999, "help": "The microsecond within a millisecond"}


@dataclass(frozen=True)
class PeriodAnchors:
    """Anchor offsets used to round datetimes up to periodic boundaries; Immutable."""

    # The anchors for a given duration unit are computed as follows:
    # yearly: Anchor(dt) = latest T where T <= dt and T == Start of January 1 of dt + anchor.yearly_* vars
    yearly_year: int = field(default=2025, metadata={"min": 1, "max": 9999, "help": "The anchor year of multi-year periods"})
    yearly_month: int = field(default=1, metadata=METADATA_MONTH)  # 1 <= x <= 12
    yearly_monthday: int = field(default=1, metadata=METADATA_DAY)  # 1 <= x <= 31
    yearly_hour: int = field(default=0, metadata=METADATA_HOUR)  # 0 <= x <= 23
    yearly_minute: int = field(default=0, metadata=METADATA_MINUTE)  # 0 <= x <= 59
    yearly_second: int = field(default=0, metadata=METADATA_SECOND)  # 0 <= x <= 59

    # monthly: Anchor(dt) = latest T <= dt at phase month (monthly_month) + anchor.monthly_* vars (day clamped; multi-month)
    monthly_month: int = field(default=1, metadata={"min": 1, "max": 12, "help": "The anchor month of multi-month periods"})
    monthly_monthday: int = field(default=1, metadata=METADATA_DAY)  # 1 <= x <= 31
    monthly_hour: int = field(default=0, metadata=METADATA_HOUR)  # 0 <= x <= 23
    monthly_minute: int = field(default=0, metadata=METADATA_MINUTE)  # 0 <= x <= 59
    monthly_second: int = field(default=0, metadata=METADATA_SECOND)  # 0 <= x <= 59

    # weekly: Anchor(dt) = latest T where T <= dt && T == Latest midnight from Sunday to Monday of dt + anchor.weekly_* vars
    weekly_weekday: int = field(default=0, metadata=METADATA_WEEKDAY)  # 0 <= x <= 6 (0=Sunday, ..., 6=Saturday)
    weekly_hour: int = field(default=0, metadata=METADATA_HOUR)  # 0 <= x <= 23
    weekly_minute: int = field(default=0, metadata=METADATA_MINUTE)  # 0 <= x <= 59
    weekly_second: int = field(default=0, metadata=METADATA_SECOND)  # 0 <= x <= 59

    # daily: Anchor(dt) = latest T where T <= dt && T == Latest midnight of dt + anchor.daily_* vars
    daily_hour: int = field(default=0, metadata=METADATA_HOUR)  # 0 <= x <= 23
    daily_minute: int = field(default=0, metadata=METADATA_MINUTE)  # 0 <= x <= 59
    daily_second: int = field(default=0, metadata=METADATA_SECOND)  # 0 <= x <= 59

    # hourly: Anchor(dt) = latest T where T <= dt && T == Latest midnight of dt + anchor.hourly_* vars
    hourly_minute: int = field(default=0, metadata=METADATA_MINUTE)  # 0 <= x <= 59
    hourly_second: int = field(default=0, metadata=METADATA_SECOND)  # 0 <= x <= 59

    # minutely: Anchor(dt) = latest T where T <= dt && T == Latest midnight of dt + anchor.minutely_* vars
    minutely_second: int = field(default=0, metadata=METADATA_SECOND)  # 0 <= x <= 59

    # secondly: Anchor(dt) = latest T where T <= dt && T == Latest midnight of dt + anchor.secondly_* vars
    secondly_millisecond: int = field(default=0, metadata=METADATA_MILLISECOND)  # 0 <= x <= 999

    # secondly: Anchor(dt) = latest T where T <= dt && T == Latest midnight of dt + anchor.millisecondly_* vars
    millisecondly_microsecond: int = field(default=0, metadata=METADATA_MICROSECOND)  # 0 <= x <= 999

    @classmethod
    def parse(cls, args: argparse.Namespace) -> PeriodAnchors:
        """Creates a ``PeriodAnchors`` instance from parsed CLI arguments."""
        kwargs: dict[str, int] = {f.name: getattr(args, f.name) for f in dataclasses.fields(cls)}
        return cls(**kwargs)

    def round_datetime_up_to_duration_multiple(self, dt: datetime, duration_amount: int, duration_unit: str) -> datetime:
        """Given a timezone-aware datetime and a duration, returns a datetime (in the same timezone) that is greater than or
        equal to dt, and rounded up (ceiled) and snapped to an anchor plus a multiple of the duration.

        The snapping is done relative to the anchors object and the rules defined therein.
        Supported units: "millisecondly", "secondly", "minutely", "hourly", "daily", "weekly", "monthly", "yearly".
        If dt is already exactly on a boundary (i.e. exactly on a multiple), it is returned unchanged.
        Examples:
        Default hourly anchor is midnight
        14:00:00, 1 hours --> 14:00:00
        14:05:01, 1 hours --> 15:00:00
        15:05:01, 1 hours --> 16:00:00
        16:05:01, 1 hours --> 17:00:00
        23:55:01, 1 hours --> 00:00:00 on the next day
        14:05:01, 2 hours --> 16:00:00
        15:00:00, 2 hours --> 16:00:00
        15:05:01, 2 hours --> 16:00:00
        16:00:00, 2 hours --> 16:00:00
        16:05:01, 2 hours --> 18:00:00
        23:55:01, 2 hours --> 00:00:00 on the next day
        """

        def add_months(dt: datetime, months: int) -> datetime:
            """Returns ``dt`` plus ``months`` with day clamped to month's end."""
            total_month: int = dt.month - 1 + months
            new_year: int = dt.year + total_month // 12
            new_month: int = total_month % 12 + 1
            last_day: int = calendar.monthrange(new_year, new_month)[1]  # last valid day of the current month
            return dt.replace(year=new_year, month=new_month, day=min(dt.day, last_day))

        def add_years(dt: datetime, years: int) -> datetime:
            """Returns ``dt`` plus ``years`` with day clamped to month's end."""
            new_year: int = dt.year + years
            last_day: int = calendar.monthrange(new_year, dt.month)[1]  # last valid day of the current month
            return dt.replace(year=new_year, day=min(dt.day, last_day))

        if duration_amount == 0:
            return dt

        period: timedelta | None = None
        anchor: datetime
        daily_base: datetime
        last_day: int
        if duration_unit == "millisecondly":
            anchor = dt.replace(hour=0, minute=0, second=0, microsecond=self.millisecondly_microsecond)
            anchor = anchor if anchor <= dt else anchor - timedelta(milliseconds=1)
            period = timedelta(milliseconds=duration_amount)

        elif duration_unit == "secondly":
            anchor = dt.replace(hour=0, minute=0, second=0, microsecond=self.secondly_millisecond * 1000)
            anchor = anchor if anchor <= dt else anchor - timedelta(seconds=1)
            period = timedelta(seconds=duration_amount)

        elif duration_unit == "minutely":
            anchor = dt.replace(second=self.minutely_second, microsecond=0)
            anchor = anchor if anchor <= dt else anchor - timedelta(minutes=1)
            period = timedelta(minutes=duration_amount)

        elif duration_unit == "hourly":
            daily_base = dt.replace(hour=0, minute=0, second=0, microsecond=0)
            anchor = daily_base + timedelta(minutes=self.hourly_minute, seconds=self.hourly_second)
            anchor = anchor if anchor <= dt else anchor - timedelta(days=1)
            period = timedelta(hours=duration_amount)

        elif duration_unit == "daily":
            daily_base = dt.replace(hour=0, minute=0, second=0, microsecond=0)
            anchor = daily_base + timedelta(hours=self.daily_hour, minutes=self.daily_minute, seconds=self.daily_second)
            anchor = anchor if anchor <= dt else anchor - timedelta(days=1)
            period = timedelta(days=duration_amount)

        elif duration_unit == "weekly":
            daily_base = dt.replace(hour=0, minute=0, second=0, microsecond=0)
            anchor = daily_base + timedelta(hours=self.weekly_hour, minutes=self.weekly_minute, seconds=self.weekly_second)
            # Convert cron weekday (0=Sunday, 1=Monday, ..., 6=Saturday) to Python's weekday (0=Monday, ..., 6=Sunday)
            target_py_weekday: int = (self.weekly_weekday - 1) % 7
            diff_days: int = (anchor.weekday() - target_py_weekday) % 7
            anchor = anchor - timedelta(days=diff_days)
            anchor = anchor if anchor <= dt else anchor - timedelta(weeks=1)
            period = timedelta(weeks=duration_amount)

        if period is not None:  # "millisecondly", "secondly", "minutely", "hourly", "daily", "weekly"
            delta: timedelta = dt - anchor
            period_micros: int = (period.days * 86400 + period.seconds) * 1_000_000 + period.microseconds
            delta_micros: int = (delta.days * 86400 + delta.seconds) * 1_000_000 + delta.microseconds
            remainder: int = delta_micros % period_micros
            if remainder == 0:
                return dt
            return dt + timedelta(microseconds=period_micros - remainder)

        elif duration_unit == "monthly":
            last_day = calendar.monthrange(dt.year, self.monthly_month)[1]  # last valid day of the anchor month
            anchor = dt.replace(  # Compute the base anchor for the anchor month ensuring the day is valid
                month=self.monthly_month,
                day=min(self.monthly_monthday, last_day),
                hour=self.monthly_hour,
                minute=self.monthly_minute,
                second=self.monthly_second,
                microsecond=0,
            )
            if anchor > dt:
                anchor = add_months(anchor, -duration_amount)
            diff_months: int = (dt.year - anchor.year) * 12 + (dt.month - anchor.month)
            anchor_boundary: datetime = add_months(anchor, duration_amount * (diff_months // duration_amount))
            if anchor_boundary < dt:
                anchor_boundary = add_months(anchor_boundary, duration_amount)
            return anchor_boundary

        elif duration_unit == "yearly":
            # Calculate the start of the cycle period that `dt` falls into.
            year_offset: int = (dt.year - self.yearly_year) % duration_amount
            period_start_year: int = dt.year - year_offset
            last_day = calendar.monthrange(period_start_year, self.yearly_month)[1]  # last valid day of the month
            anchor = dt.replace(
                year=period_start_year,
                month=self.yearly_month,
                day=min(self.yearly_monthday, last_day),
                hour=self.yearly_hour,
                minute=self.yearly_minute,
                second=self.yearly_second,
                microsecond=0,
            )
            if anchor < dt:
                return add_years(anchor, duration_amount)
            return anchor

        else:
            raise ValueError(f"Unsupported duration unit: {duration_unit}")
